@@ -1,23 +1,92 @@
 #
-#   Copyright 2007-2019 by the individuals mentioned in the source code history
+#   Copyright 2007-2021 by the individuals mentioned in the source code history
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
 #   you may not use this file except in compliance with the License.
 #   You may obtain a copy of the License at
-# 
+#
 #        http://www.apache.org/licenses/LICENSE-2.0
-# 
+#
 #   Unless required by applicable law or agreed to in writing, software
 #   distributed under the License is distributed on an "AS IS" BASIS,
 #   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 #   See the License for the specific language governing permissions and
 #   limitations under the License.
 
+##' The MxCompare Class
+##'
+##' @aliases
+##' $,MxCompare-method
+##' [,MxCompare,ANY,ANY,ANY-method
+##' as.data.frame.MxCompare
+##' print,MxCompare-method
+##' show,MxCompare-method
+##'
+##' @details
+##' This is an internal class structure.  You should not use it directly.
+##' Use \code{\link{mxCompare}} instead.
+setClass(Class =  "MxCompare",
+	representation = representation(
+		results = 'data.frame',
+		base = 'list',
+		comparison = 'list'
+	)
+)
+
+# Initialize method for the new() function
+setMethod("initialize", "MxCompare",
+	function(.Object, results, base, comparison){
+		.Object@results <- results
+		.Object@base <- base
+		.Object@comparison <- comparison
+		return(.Object)
+	}
+)
+
+cmpMlNames <- c('base', 'comparison', 'ep', 'minus2LL', 'df', 'AIC', 'diffLL', 'diffdf', 'p')
+cmpWlsNames <- c('base', 'comparison', 'ep', 'chisq', 'df', 'AIC', 'SBchisq', 'diffdf', 'p')
+cmpCommonNames <- intersect(cmpMlNames, cmpWlsNames)
+cmpAbstractNames <- c('fit', 'fitUnits', 'diffFit')
+
+displayCompare <- function(obj){
+	rfu <- obj@results[1, c('fitUnits')]
+	pnames <- if(rfu == "r'Wr"){cmpWlsNames} else if(rfu == '-2lnL'){cmpMlNames} else {stop("OpenMx can't handle the fit units you have.")}
+	print(obj@results[, pnames, drop=FALSE])
+}
+setMethod("print", "MxCompare", function(x,...) { displayCompare(x) })
+setMethod("show", "MxCompare", function(object) { displayCompare(object) })
+
+setMethod("$", "MxCompare",
+          function(x, name){x@results[ , name]}
+)
+
+.DollarNames.MxCompare <- function(x, pattern){
+	if(missing(pattern)){
+		pattern <- ''
+	}
+	output <- names(x@results)
+	output <- gsub("(\\w+\\W+.*)", "'\\1'", output)
+	return(grep(pattern, output, value=TRUE))
+}
+
+setMethod("[", "MxCompare",
+	function (x, i, j, drop=if (missing(i)) TRUE else !missing(j) && length(j) == 1){
+		rfu <- x@results[1, c('fitUnits')]
+		pnames <- if(rfu == "r'Wr"){cmpWlsNames} else if(rfu == '-2lnL'){cmpMlNames} else {stop("OpenMx can't handle the fit units you have.")}
+		tmp <- x@results[, c(pnames, setdiff(pnames, cmpCommonNames), cmpAbstractNames), drop=FALSE]
+		tmp[i, j, drop]
+	}
+)
+
+as.data.frame.MxCompare <- function(x, row.names = NULL, optional = FALSE, ...){
+	return(data.frame(x@results, row.names=row.names))
+}
+
 mxCompare <- function(base, comparison, ..., all = FALSE,
 		      boot=FALSE, replications=400, previousRun=NULL, checkHess=FALSE) {
 	prohibitDotdotdot(list(...))
 	if (missing(base)) {
-		stop("'base' argument be a MxModel object or list of MxModel objects")	
+		stop("'base' argument be a MxModel object or list of MxModel objects")
 	}
 	if (is.list(base)) {
 		base <- unlist(base)
@@ -27,7 +96,7 @@ mxCompare <- function(base, comparison, ..., all = FALSE,
 	if(!all(sapply(base, is, "MxModel"))) {
 		stop("The 'base' argument must consist of MxModel objects")
 	}
-	
+
 	if (missing(comparison)) {
 		# no comparison models, just make a dummy list to feed to showFitStatistics
 		comparison <- list()
@@ -41,11 +110,11 @@ mxCompare <- function(base, comparison, ..., all = FALSE,
 			stop("The 'comparison' argument must consist of MxModel objects")
 		}
 	}
-		
+
 	if (missing(checkHess)) checkHess <- as.logical(NA)
 	if (missing(boot) && (!missing(replications) || !missing(previousRun))) boot <- TRUE
 	resultsTable <- showFitStatistics(base, comparison, all, boot, replications, previousRun, checkHess)
-	return(resultsTable)
+	return(new("MxCompare", resultsTable, base, comparison))
 }
 
 mxCompareMatrix <- function(models,
@@ -54,7 +123,7 @@ mxCompareMatrix <- function(models,
 			    boot=FALSE, replications=400, previousRun=NULL, checkHess=FALSE,
 			    wholeTable=FALSE) {
 	prohibitDotdotdot(list(...))
-	
+
 	if (missing(checkHess)) checkHess <- as.logical(NA)
 	if (missing(boot) && (!missing(replications) || !missing(previousRun))) boot <- TRUE
 
@@ -90,7 +159,7 @@ anova.MxModel <- function(object, ...) {
 		replications <- args[[ap]]
 		args <- args[-ap]
 	}
-	
+
 	previousRun <- NULL
 	ap <- match('previousRun', names(args))
 	if (!is.na(ap)) {
@@ -122,9 +191,9 @@ anova.MxModel <- function(object, ...) {
 	}
 	models <- c(list(object), comparison)
 	summaries <- lapply(models, summary)
-	stats <- data.frame(m2ll=unlist(Map(function(x) x$Minus2LogLikelihood, summaries)),
+	stats <- data.frame(fit=unlist(Map(function(x) x$fit, summaries)),
 			    df  =unlist(Map(function(x) x$degreesOfFreedom, summaries)))
-	mo <- with(stats, order(df, m2ll))
+	mo <- with(stats, order(df, fit))
 	models <- models[mo]
 	stats <- stats[mo,]
 
@@ -132,7 +201,7 @@ anova.MxModel <- function(object, ...) {
 	if (boot) {
 		bootTodo <- list()
 		for (rx in 2:length(models)) {
-			if (stats[rx-1,'m2ll'] >= stats[rx,'m2ll'] ||
+			if (stats[rx-1,'fit'] >= stats[rx,'fit'] ||
 			    stats[rx-1,'df'] >= stats[rx,'df']) next
 			nChar <- as.character(rx)
 			bootTodo[[ nChar ]] <- rx-1
@@ -143,11 +212,11 @@ anova.MxModel <- function(object, ...) {
 
 	result <- list()
 
-	result[[ length(result) + 1L ]] <- 
+	result[[ length(result) + 1L ]] <-
 		collectBaseStatistics(newEmptyCompareRow(), models[[1]])
 
 	for (i in 2:length(models)) {
-		if (stats[i-1,'m2ll'] > stats[i,'m2ll']) {
+		if ( stats[i-1,'fit'] > stats[i,'fit'] ) {
 			result[[ length(result) + 1L ]] <-
 				collectBaseStatistics(newEmptyCompareRow(), models[[i]])
 			next
@@ -161,6 +230,11 @@ anova.MxModel <- function(object, ...) {
 	if (boot) {
 		attr(ret, "bootData") <- bootData
 	}
+
+	rfu <- ret[1, c('fitUnits')]
+	pnames <- if(rfu == "r'Wr"){cmpWlsNames} else if(rfu == '-2lnL'){cmpMlNames} else {names(ret)}
+	ret <- ret[, pnames, drop=FALSE]
+
 	ret
 }
 
@@ -173,13 +247,18 @@ newEmptyCompareRow <- function() {
 		   df=as.numeric(NA),
 		   AIC=as.numeric(NA),
 		   diffLL=as.numeric(NA),
-		   diffdf=as.numeric(NA), 
-		   p=as.numeric(NA))
+		   diffdf=as.numeric(NA),
+		   p=as.numeric(NA),
+		   fit=as.numeric(NA),
+		   fitUnits=as.character(NA),
+		   diffFit=as.numeric(NA),
+		   chisq=as.numeric(NA),
+		   SBchisq=as.numeric(NA))
 }
 
 iterateNestedModels <- function(models, boot, replications, previousRun, checkHess) {
 	result <- list()
-	
+
 	summaries <- lapply(models, summary)
 
 	bd <- NULL
@@ -225,36 +304,36 @@ iterateNestedModels <- function(models, boot, replications, previousRun, checkHe
 }
 
 assertIsRawData <- function(model) {
-  type <- model$data$type
-  if (type == 'raw') return()
-  stop(paste("Model", omxQuotes(model$name), "contains", omxQuotes(type),
-	     "data. Only type='raw' data is supported by",
-	     "mxPowerSearch(..., method='empirical')"))
+	type <- model$data$type
+	if (type == 'raw') return()
+	stop(paste("Model", omxQuotes(model$name), "contains", omxQuotes(type),
+		"data. Only type='raw' data is supported by",
+		"mxPowerSearch(..., method='empirical')"))
 }
 
 loadDataIntoModel <- function(model, dataList, assertRaw=FALSE) {
-  for (modelName in names(dataList)) {
-    dataobj <- mxData(dataList[[modelName]], type='raw')
-    if (modelName == model$name) {
-      if (assertRaw) assertIsRawData(model)
-      model <- mxModel(model, dataobj)
-    } else {
-      if (assertRaw) assertIsRawData(model[[modelName]])
-      model <- mxModel(model, mxModel(model[[modelName]], dataobj))
-    }
-  }
-  model
+	for (modelName in names(dataList)) {
+		dataobj <- mxData(dataList[[modelName]], type='raw')
+		if (modelName == model$name) {
+			if (assertRaw) assertIsRawData(model)
+			model <- mxModel(model, dataobj)
+		} else {
+			if (assertRaw) assertIsRawData(model[[modelName]])
+			model <- mxModel(model, mxModel(model[[modelName]], dataobj))
+		}
+	}
+	model
 }
 
 setupBootData <- function(nullHyp, comparison, todo,
 			  replications, previousRun) {
-  
+
   # pre-check data compatibility of nullHyp and comparison? TODO
 
-	if (!is.null(previousRun)) previousRun <- attr(previousRun, "bootData")
+	if (!is.null(previousRun)) previousRun <- attr(previousRun@results, "bootData")
 
   seedVec <- as.integer(runif(replications, min = -2e9, max=2e9))
-  
+
   bootData <- list()
   for (i in names(todo)) {
     nullHypData <- data.frame(seed=seedVec, fit=NA,
@@ -321,11 +400,11 @@ fillBootData <- function(nullHyp, comparison, todo, bootData, checkHess) {
 			  nullHypData[repl, 'fit'] <- null1$output$fit
 			  nullHypData[repl, names(coef(null1))] <- coef(null1)
 			  nullHypData[repl, 'statusCode'] <- as.statusCode(null1$output$status$code)
-			  
+
 			  bootData[[ i ]] <- nullHypData
 
 			  topDataIndex <- match(null1$name, names(simData))
-			  
+
 			  for (j in todo[[i]]) {
 				  cmpData <- bootData[[ paste(i,j,sep=":") ]]
 				  if (!is.na(cmpData[repl, 'fit'])) next
@@ -339,7 +418,7 @@ fillBootData <- function(nullHyp, comparison, todo, bootData, checkHess) {
 				  cmpData[repl, 'fit'] <- cmp1$output$fit
 				  cmpData[repl, names(coef(cmp1))] <- coef(cmp1)
 				  cmpData[repl, 'statusCode'] <- as.statusCode(cmp1$output$status$code)
-				  
+
 				  bootData[[ paste(i,j,sep=":") ]] <- cmpData
 			  }
 		  }
@@ -389,63 +468,81 @@ showFitStatistics <- function(base, compare, all, boot, replications, previousRu
 		bootData <- fillBootData(compare, base, bootTodo, bootData, checkHess)
 	}
 
-    statistics <- list()
-    if (all) {
-        for (i in seq_along(base)) {
-            nextBase <- base[[i]]
-            statistics[[length(statistics) + 1]] <-
-		    collectBaseStatistics(newEmptyCompareRow(), nextBase)
-            for (j in seq_along(compare)) {
-                nextCompare <- compare[[j]]
-		boot1 <- extractLRTBootstrapPair(bootData, i, j)
-                statistics[[length(statistics) + 1]] <-
-			collectStatistics(newEmptyCompareRow(), nextBase, nextCompare, boot1)
-            }
-        }
-    }
-    else {
-		if(length(compare)==0){
-	        for (i in seq_along(base)) {
+	statistics <- list()
+	if (all) {
+		for (i in seq_along(base)) {
+			nextBase <- base[[i]]
 			statistics[[length(statistics) + 1]] <-
-				collectBaseStatistics(newEmptyCompareRow(), base[[i]])
-	        }
+			collectBaseStatistics(newEmptyCompareRow(), nextBase)
+			for (j in seq_along(compare)) {
+				nextCompare <- compare[[j]]
+				boot1 <- extractLRTBootstrapPair(bootData, i, j)
+				statistics[[length(statistics) + 1]] <-
+					collectStatistics(newEmptyCompareRow(), nextBase, nextCompare, boot1)
+			}
+		}
+	} else {
+		if(length(compare)==0){
+			for (i in seq_along(base)) {
+				statistics[[length(statistics) + 1]] <-
+					collectBaseStatistics(newEmptyCompareRow(), base[[i]])
+			}
 		} else {
-	        maxLength <- max(length(base), length(compare))
-	        previousBaseIndex <- -1
-	        for (i in 1:maxLength) {
-	            nextBaseIndex <- (i - 1)%%length(base) + 1
-	            nextCompareIndex <- (i - 1)%%length(compare) + 1
-	            nextBase <- base[[nextBaseIndex]]
-	            nextCompare <- compare[[nextCompareIndex]]
-	            if (previousBaseIndex != nextBaseIndex) {
-	                statistics[[length(statistics) + 1]] <-
-				collectBaseStatistics(newEmptyCompareRow(), nextBase)
-	            }
-		    boot1 <- extractLRTBootstrapPair(bootData, nextBaseIndex, nextCompareIndex)
-	            statistics[[length(statistics) + 1]] <-
-			    collectStatistics(newEmptyCompareRow(), nextBase, nextCompare, boot1)
-	            previousBaseIndex <- nextBaseIndex
-	        }
-		}		
-    }
+			maxLength <- max(length(base), length(compare))
+			previousBaseIndex <- -1
+			for (i in 1:maxLength) {
+				nextBaseIndex <- (i - 1)%%length(base) + 1
+				nextCompareIndex <- (i - 1)%%length(compare) + 1
+				nextBase <- base[[nextBaseIndex]]
+				nextCompare <- compare[[nextCompareIndex]]
+				if (previousBaseIndex != nextBaseIndex) {
+					statistics[[length(statistics) + 1]] <-
+						collectBaseStatistics(newEmptyCompareRow(), nextBase)
+				}
+				boot1 <- extractLRTBootstrapPair(bootData, nextBaseIndex, nextCompareIndex)
+				statistics[[length(statistics) + 1]] <-
+					collectStatistics(newEmptyCompareRow(), nextBase, nextCompare, boot1)
+				previousBaseIndex <- nextBaseIndex
+			}
+		}
+	}
 	statistics <- do.call(rbind, statistics)
 	if (boot) {
 		attr(statistics, "bootData") <- bootData
 	}
-    return(statistics)
+	return(statistics)
 }
 
 collectBaseStatistics <- function(row, ref) {
 	refSummary <- summary(ref)
 	row[,'base'] <- refSummary$modelName
-	row[,c('ep','minus2LL','df','AIC')] <-
+	row[,c('ep', 'minus2LL', 'AIC')] <-
 		c(refSummary$estimatedParameters,
 		  refSummary$Minus2LogLikelihood,
-		  refSummary$degreesOfFreedom,
-		  refSummary$AIC.Mx)
+		  AIC(ref))
+	rfu <- ref$output$fitUnits
+	row[, 'fitUnits'] <- rfu
+	row[, 'chisq'] <- ifelse(!is.null(ref$output$chi), ref$output$chi, NA)
+	row[, c('fit', 'df')] <- if(rfu == "r'Wr"){
+			c(ref$output$chi, ref$output$chiDoF)
+		} else if(rfu == '-2lnL'){
+			c(refSummary$Minus2LogLikelihood, refSummary$degreesOfFreedom)
+		} else {
+			stop("Can't compare models with these fit units: ", rfu)
+		}
 	row
 }
 
+# Satorra-Bentler Chi-squared fit (Fit Satorra Bentler = FSB)
+fsb <- function(chi0, chi1, df0, df1, chim0, chim1){
+	sb <- (chi0 - chi1)*(df0 - df1) / (df0*chi0/chim0 - df1*chi1/chim1)
+	return(list(chi=sb, df=df0-df1))
+}
+
+fsb_helper <- function(model0, model1){
+	fsb(model0$output$chi, model1$output$chi, model0$output$chiDoF, model1$output$chiDoF, model0$output$chiM, model1$output$chiM)
+}
+
 collectStatistics <- function(otherStats, ref, other, bootPair) {
 	otherStats <- collectBaseStatistics(otherStats, other)
 	collectStatistics1(otherStats, ref, other, bootPair)
@@ -456,7 +553,7 @@ collectStatistics1 <- function(otherStats, ref, other, bootPair) {
 	assertModelRunAndFresh(other)
 	refSummary <- summary(ref)
 	otherSummary <- summary(other)
-	
+
 	#Check for validity of the comparison ###
 	rfu <- ref$output$fitUnits #<--NULL if output slot is empty.
 	if(!length(rfu)){
@@ -480,21 +577,25 @@ collectStatistics1 <- function(otherStats, ref, other, bootPair) {
 	if(length(ogfe)){ogfe <- paste(ogfe$name,collapse=",")}
 	if( length(rgfe)!=length(ogfe) || (length(rgfe) && length(ogfe) && rgfe!=ogfe) ){
 		#This is a warning, not an error, because it's possible that the user is indeed using the same covariates in both models, but with
-		#different column names.  (If one of the models hasn't been run yet, GREMLFixEffList() will return NULL, but the fit value will 
+		#different column names.  (If one of the models hasn't been run yet, GREMLFixEffList() will return NULL, but the fit value will
 		#be NA, so the output for the comparison won't even look valid):
 		warning(paste("the names of the fixed effects in MxModels '",ref$name,"' and '",other$name,"' do not match; comparison of REML fit values is only valid for models that use the same covariates",sep=""))
 	}
 	#End validity checks
-	
+
 	otherStats[,c('base','comparison')] <-
 		c(refSummary$modelName,
 		  otherSummary$modelName)
 	otherStats[,c('diffLL','diffdf')] <-
 		c(otherSummary$Minus2LogLikelihood - refSummary$Minus2LogLikelihood,
 		  otherSummary$degreesOfFreedom - refSummary$degreesOfFreedom)
+	# TODO ? need to adjust DoF for Chi-sq with WLS?  In Theory these should always be the same.
+	otherStats[,c('SBchisq')] <- fsb_helper(other, ref)$chi
 
-  diffdf <- otherStats[['diffdf']]
-  diffdf <- diffdf[!is.na(diffdf)]
+	otherStats[,c('diffFit')] <- if(rfu[[1]] == "r'Wr"){otherStats[,'SBchisq']} else if(rfu[[1]] == '-2lnL'){otherStats[,'diffLL']} else {stop("Unknown fitUnits")}
+
+	diffdf <- otherStats[['diffdf']]
+	diffdf <- diffdf[!is.na(diffdf)]
 	if (length(diffdf) && any(diffdf < 0)) {
 		msg <- paste("Model", omxQuotes(refSummary$modelName), "has more degrees of freedom than",
 			     otherSummary$modelName, "which means that the models need to be",
@@ -502,23 +603,20 @@ collectStatistics1 <- function(otherStats, ref, other, bootPair) {
 		warning(msg)
 	} else {
 		if (is.null(bootPair)) {
-			otherStats[['p']] <- pchisq(otherStats[['diffLL']], otherStats[['diffdf']], lower.tail=FALSE)
+			otherStats[['p']] <- pchisq(otherStats[['diffFit']], otherStats[['diffdf']], lower.tail=FALSE)
 		} else {
 			baseData <- bootPair[[1]] # null hypothesis
 			cmpData <- bootPair[[2]]
-			
+
 			if (any(baseData[,'seed'] != cmpData[,'seed'])) {
 				stop("Some seeds do not match")
 			}
-			
+
 			# If optimization goes wrong, the null model could fit better than
 			# the alternative model. We exclude these replications.
 			mask <- (baseData[,'statusCode'] %in% mxOption(other, "Status OK") &
 				 cmpData[,'statusCode'] %in% mxOption(ref, "Status OK") &
 				 baseData[,'fit'] - cmpData[,'fit'] > 0)
-			if (sum(mask) < 3) {
-				stop(paste("Fewer than 3 replications are available."))
-			}
 			if (sum(mask) < .95*length(mask)) {
 				pct <- round(100*sum(mask) / length(mask))
 				warning(paste0("Only ",pct,"% of the bootstrap replications ",
@@ -552,7 +650,7 @@ mxParametricBootstrap <- function(nullModel, labels,
   bootData <- data.frame(seed=seedVec, fit=NA,
                          statusCode=as.statusCode(NA))
   for (par in c(names(coef(nullModel)), labels)) bootData[[par]] <- NA
-  
+
 	if (!is.null(previousRun)) {
 		previousRun <- attr(previousRun, "bootData")
 		if (ncol(previousRun) != ncol(bootData) || any(colnames(previousRun) != colnames(bootData))) {
@@ -573,15 +671,12 @@ mxParametricBootstrap <- function(nullModel, labels,
     bootData[repl, names(coef(base1))] <- coef(base1)
     bootData[repl, 'statusCode'] <- as.statusCode(base1$output$status$code)
   }
-  
+
   nullParam <- omxGetParameters(nullModel, free=NA)[labels]
   model <- omxSetParameters(nullModel, labels=labels, free=TRUE)
   model <- mxRun(model, silent = TRUE, suppressWarnings = TRUE)
-  
+
   mask <- bootData[,'statusCode'] %in% mxOption(nullModel, "Status OK")
-  if (sum(mask) < 3) {
-    stop(paste("Fewer than 3 replications are available."))
-  }
   if (sum(mask) < .95*length(mask)) {
     pct <- round(100*sum(mask) / length(mask))
     warning(paste0("Only ",pct,"% of the bootstrap replications ",
@@ -589,7 +684,7 @@ mxParametricBootstrap <- function(nullModel, labels,
                    replications,
                    " replications requested"), call.=FALSE)
   }
-  
+
   est <- coef(model)[labels]
   pval <- rep(NA, length(labels))
   for (lx in seq_along(labels)) {
@@ -602,7 +697,7 @@ mxParametricBootstrap <- function(nullModel, labels,
     }
   }
   pval <- pval / sum(mask)
-  
+
   pzero <- pval == 0
   pval[pzero] <- 1/sum(mask)
   pval <- p.adjust(pval, method=correction)
@@ -624,52 +719,52 @@ mxParametricBootstrap <- function(nullModel, labels,
 }
 
 totalSampleSize <- function(model, default=100L) {
-  sizes <- sapply(extractData(model), function(mxd) {
-    if (mxd@type == 'raw') {
-      nrow(mxd@observed)
-    } else {
-      mxd@numObs
-    }
-  })
-  if (length(sizes) == 0) sizes <- default
-  sum(sizes)
+	sizes <- sapply(extractData(model), function(mxd) {
+		if (mxd@type == 'raw') {
+			nrow(mxd@observed)
+		} else {
+			mxd@numObs
+		}
+	})
+	if (length(sizes) == 0) sizes <- default
+	sum(sizes)
 }
 
 fitPowerModel <- function(rx, result) {
-  # rx is which(is.na(result$reject))[1] - 1L
-  result <- result[!is.na(result$x),]
-  algRle <- rle(result$alg)
-  if (algRle$values[1] == 'init') {
-    # skip outliers from early probe sequence
-    newFirst <- max(1, algRle$lengths[1] - 3L)
-    result <- result[newFirst:nrow(result),]
-  }
-  m1 <- suppressWarnings(glm(reject ~ x, data = result, family = binomial))
-  if (all(!is.na(coef(m1))) && summary(m1)$coefficients['x','Pr(>|z|)'] < .25) {
-    curX <- as.numeric(((rx %% 3) - coef(m1)[1]) / coef(m1)[2])
-    alg <- '2p'
-  } else {
-    m2 <- glm(reject ~ 1, data = result, family = binomial)
-    from <- max(nrow(result)-9, 1)
-    curX <- mean(result$x[from:nrow(result)]) * ifelse(coef(m2)[1] < 0, 1.1, 0.9)
-    alg <- '1p'
-  }
-    # Can only consider one-sided hypotheses
-  if (sign(curX) != sign(result[1,'x'])) curX <- 0
-  list(curX=curX, m1=m1, alg=alg)
+	# rx is which(is.na(result$reject))[1] - 1L
+	result <- result[!is.na(result$x),]
+	algRle <- rle(result$alg)
+	if (algRle$values[1] == 'init') {
+		# skip outliers from early probe sequence
+		newFirst <- max(1, algRle$lengths[1] - 3L)
+		result <- result[newFirst:nrow(result),]
+	}
+	m1 <- suppressWarnings(glm(reject ~ x, data = result, family = binomial))
+	if (all(!is.na(coef(m1))) && summary(m1)$coefficients['x','Pr(>|z|)'] < .25) {
+		curX <- as.numeric(((rx %% 3) - coef(m1)[1]) / coef(m1)[2])
+		alg <- '2p'
+	} else {
+		m2 <- glm(reject ~ 1, data = result, family = binomial)
+		from <- max(nrow(result)-9, 1)
+		curX <- mean(result$x[from:nrow(result)]) * ifelse(coef(m2)[1] < 0, 1.1, 0.9)
+		alg <- '1p'
+	}
+	# Can only consider one-sided hypotheses
+	if (sign(curX) != sign(result[1,'x'])) curX <- 0
+	list(curX=curX, m1=m1, alg=alg)
 }
 
 validateSigLevel <- function(sl) {
-  if (length(sl) > 1) {
-    stop(paste("Can only evaluate one sig.level at a time.",
-	       "To evaluate power across a range of sig.levels",
-	       "you need to call mxPower in a loop and accumulate",
-	       "estimates that way"))
-  } else if (length(sl) == 0) {
-    stop("At what sig.level?")
-  } else if (sl <= 0 || sl >= 1) {
-    stop("sig.level must be between 0 and 1")
-  }
+	if (length(sl) > 1) {
+		stop(paste("Can only evaluate one sig.level at a time.",
+			"To evaluate power across a range of sig.levels",
+			"you need to call mxPower in a loop and accumulate",
+			"estimates that way"))
+	} else if (length(sl) == 0) {
+		stop("At what sig.level?")
+	} else if (sl <= 0 || sl >= 1) {
+		stop("sig.level must be between 0 and 1")
+	}
 }
 
 mxPowerSearch <- function(trueModel, falseModel, n=NULL, sig.level=0.05, ...,
@@ -722,7 +817,7 @@ mxPowerSearch <- function(trueModel, falseModel, n=NULL, sig.level=0.05, ...,
     if (avgNcp == 0.0) stop("falseModel and trueModel are identical?")
     # is diffdf>1 ever a good approx? TODO
     diffdf <- summary(falseModel)[['degreesOfFreedom']] - summary(trueModel)[['degreesOfFreedom']]
-    
+
     if (is.null(grid)) {
       width <- 2.75/avgNcp
       center <- 1.15*qchisq(1 - sig.level, diffdf)/avgNcp
@@ -741,7 +836,7 @@ mxPowerSearch <- function(trueModel, falseModel, n=NULL, sig.level=0.05, ...,
     colnames(out)[1] <- 'N'
     return(out)
   }
-  
+
   interest <- setdiff(names(coef(trueModel)), names(coef(falseModel)))
   if (!is.null(n) && length(interest) != 1) {
     stop(paste("Specify only 1 parameter (not", omxQuotes(interest),
@@ -827,16 +922,16 @@ mxPowerSearch <- function(trueModel, falseModel, n=NULL, sig.level=0.05, ...,
       simData <- list(simData)
       names(simData) <- trueModel$name
     }
-    
+
     true1  <- loadDataIntoModel(trueModel,  simData, assertRaw=rx==1)
     true1  <- mxRun(true1,  silent=TRUE, suppressWarnings = TRUE)
     # complain about parameters at box constraints TODO
-    
+
     topDataIndex <- match(trueModel$name, names(simData))
     names(simData)[topDataIndex] <- falseModel$name
     false1 <- loadDataIntoModel(falseModel, simData)
     false1 <- mxRun(false1, silent=TRUE, suppressWarnings = TRUE)
-    
+
     if (statistic == 'LRT') {
 	    cmp1 <- mxCompare(true1, false1)
 	    pval <- cmp1[2,'p']
@@ -872,6 +967,7 @@ mxPowerSearch <- function(trueModel, falseModel, n=NULL, sig.level=0.05, ...,
       curX <- pm$curX
       alg <- pm$alg
     }
+    if (is.null(n)) curX <- max(curX, 2/origSampleSize) # lower bound for N
   }
   if (!silent) imxReportProgress('', prevProgressLen)
   if (is.null(m1)) stop("Logistic model failed to converge")
@@ -937,7 +1033,7 @@ mxPower <- function(trueModel, falseModel, n=NULL, sig.level=0.05, power=0.8, ..
                                       checkHess=checkHess))
     return(got)
   }
-  
+
   origSampleSize <- totalSampleSize(trueModel)
   method <- match.arg(method)
   statistic <- match.arg(statistic)
@@ -954,7 +1050,7 @@ mxPower <- function(trueModel, falseModel, n=NULL, sig.level=0.05, power=0.8, ..
       result <- data.frame(seed=as.integer(runif(probes, min = -2e9, max=2e9)),
         reject=NA, mseTrue=NA,
         statusTrue=as.statusCode(NA), statusFalse=as.statusCode(NA))
-      
+
       trueModel <- ProcessCheckHess(trueModel, checkHess)
       falseModel <- ProcessCheckHess(falseModel, checkHess)
 
@@ -975,11 +1071,11 @@ mxPower <- function(trueModel, falseModel, n=NULL, sig.level=0.05, power=0.8, ..
           simData <- list(simData)
           names(simData) <- trueModel$name
         }
-        
+
         true1  <- loadDataIntoModel(trueModel,  simData)
         true1  <- mxRun(true1,  silent=TRUE, suppressWarnings = TRUE)
         # complain about parameters at box constraints TODO
-        
+
         topDataIndex <- match(trueModel$name, names(simData))
         names(simData)[topDataIndex] <- falseModel$name
         false1 <- loadDataIntoModel(falseModel, simData)
