@@ -1,31 +1,18 @@
 #
-#   Copyright 2007-2019 by the individuals mentioned in the source code history
+#   Copyright 2007-2021 by the individuals mentioned in the source code history
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
 #   you may not use this file except in compliance with the License.
 #   You may obtain a copy of the License at
-# 
+#
 #        http://www.apache.org/licenses/LICENSE-2.0
-# 
+#
 #   Unless required by applicable law or agreed to in writing, software
 #   distributed under the License is distributed on an "AS IS" BASIS,
 #   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 #   See the License for the specific language governing permissions and
 #   limitations under the License.
 
-##' @name MxDataFrameOrMatrix-class
-##' @rdname MxDataFrameOrMatrix-class
-##' @title MxDataFrameOrMatrix
-##'
-##' @description
-##' Internal class that is the union of data.frame and matrix.
-##'
-##' @aliases
-##' MxDataFrameOrMatrix
-##' @details
-##' Not to be used.
-setClassUnion("MxDataFrameOrMatrix", c("data.frame", "matrix"))
-
 setClass(Class = "NonNullData",
 	representation = representation(
 		verbose = "integer"),
@@ -47,7 +34,7 @@ setClass(Class = "NonNullData",
 setClass(Class = "MxDataStatic",
 	 contains = "NonNullData",
 	 representation = representation(
-		observed = "MxDataFrameOrMatrix",
+		observed = "MxOptionalDataFrameOrMatrix",
 		means  = "matrix",
 		type   = "character",
 		numObs = "numeric",
@@ -61,9 +48,11 @@ setClass(Class = "MxDataStatic",
 	     frequency = "MxCharOrNumber",
 	     minVariance = "numeric",
 	   algebra = "MxOptionalCharOrNumber",
-    warnNPDacov = "logical",
+    warnNPDuseWeight = "logical",
     exoFree = "MxOptionalMatrix",
-    naAction = "character"))
+    naAction = "character",
+    fitTolerance = "numeric",
+    gradientTolerance = "numeric"))
 
 setClass(Class = "MxDataDynamic",
 	 contains = "NonNullData",
@@ -77,7 +66,8 @@ setClassUnion("MxData", c("NULL", "MxDataStatic", "MxDataDynamic"))
 setMethod("initialize", "MxDataStatic",
 	  function(.Object, observed, means, type, numObs, observedStats,
 		   sort, primaryKey, weight, frequency, verbose, .parallel, .noExoOptimize,
-		   minVariance, algebra, warnNPDacov, exoFree, naAction) {
+		   minVariance, algebra, warnNPDuseWeight, exoFree, naAction,
+       fitTolerance, gradientTolerance) {
 		.Object@observed <- observed
 		.Object@means <- means
 		.Object@type <- type
@@ -104,9 +94,11 @@ setMethod("initialize", "MxDataStatic",
 		.Object@verbose <- verbose
 		.Object@minVariance <- minVariance
 		.Object@algebra <- algebra
-		.Object@warnNPDacov <- warnNPDacov
+		.Object@warnNPDuseWeight <- warnNPDuseWeight
 		.Object@exoFree <- exoFree
 		.Object@naAction <- naAction
+    .Object@fitTolerance <- fitTolerance
+    .Object@gradientTolerance <- gradientTolerance
 		return(.Object)
 	}
 )
@@ -144,6 +136,17 @@ nrowMxData <- function(mxd) {
     }
 }
 
+observedDataNames <- function(mxd) {
+  if (mxd@type == 'none') {
+    obs <- mxd@observedStats[['cov']]
+    if (!is.null(obs)) {
+      colnames(obs)
+    } else c()
+  } else {
+    colnames(mxd@observed)
+  }
+}
+
 ##' Create dynamic data
 ##'
 ##' @param type type of data
@@ -161,33 +164,40 @@ mxDataDynamic <- function(type, ..., expectation, verbose=0L) {
 	return(new("MxDataDynamic", type, expectation, verbose))
 }
 
-mxData <- function(observed, type, means = NA, numObs = NA, acov=NA, fullWeight=NA,
+mxData <- function(observed=NULL, type="none",
+                   means = NA, numObs = NA, acov=NA, fullWeight=NA,
 		   thresholds=NA, ...,
 		   observedStats=NA, sort=NA, primaryKey = as.character(NA), weight = as.character(NA),
 		   frequency = as.character(NA), verbose=0L, .parallel=TRUE, .noExoOptimize=TRUE,
 		   minVariance=sqrt(.Machine$double.eps), algebra=c(),
-		   warnNPDacov=TRUE, exoFree=NULL, naAction=c("pass","fail","omit","exclude")) {
+		   warnNPDacov=TRUE, warnNPDuseWeight=TRUE,
+       exoFree=NULL, naAction=c("pass","fail","omit","exclude"),
+       fitTolerance=sqrt(as.numeric(mxOption(key="Optimality tolerance"))),
+       gradientTolerance=1e-2) {
   prohibitDotdotdot(list(...))
 	if (length(means) == 1 && is.na(means)) means <- as.numeric(NA)
-	if (missing(observed) || !is(observed, "MxDataFrameOrMatrix")) {
-		stop("Observed argument is neither a data frame nor a matrix")
-	}
 	dups <- duplicated(colnames(observed))
 	if (any(dups)) {
 		stop(paste("Column names must be unique. Duplicated:",
 			   omxQuotes(colnames(observed)[dups])))
 	}
-	if (missing(type) || (!is.character(type)) || (length(type) > 1) || 
-		is.na(match(type, imxDataTypes))) {
+  if (missing(observed) && type != 'none') {
+    # We use 'none' instead of NA to avoid adding is.null checks everywhere
+    stop(paste0("No observed data but type='", type,"' (should be 'none')"))
+  }
+	if (!missing(observed) && (missing(type) || (!is.character(type)) || (length(type) > 1) ||
+		is.na(match(type, imxDataTypes)))) {
 			numTypes = length(imxDataTypes)
-			stop(paste("Type must be set to one of: ", "'", 
+			stop(paste("Type must be set to one of: ", "'",
 			paste(imxDataTypes[1:(numTypes-1)], collapse="' '"),
 			"' or '", imxDataTypes[numTypes], "'", sep=""))
 	}
   naAction <- match.arg(naAction)
 	if (type == "acov") {
-		return(legacyMxData(observed, type="acov", means=means, numObs=numObs, 
-			acov=acov, fullWeight=fullWeight, thresholds=thresholds))
+		obj <- legacyMxData(observed, type="acov", means=means, numObs=numObs,
+                        acov=acov, fullWeight=fullWeight, thresholds=thresholds)
+    obj@verbose <- as.integer(verbose)
+    return(obj)
 	}
 	if (missing(observedStats)) {
 		observedStats <- list()
@@ -205,18 +215,20 @@ mxData <- function(observed, type, means = NA, numObs = NA, acov=NA, fullWeight=
 	if ((!is.vector(means) && !(prod(dim(means)) == length(means))) || !is.numeric(means)) {
 		stop("Means argument must be of numeric vector type")
 	}
-	if (type != "raw" && is.na(numObs)) {
-		stop("Number of observations must be specified for non-raw data, i.e., add numObs=XXX to mxData()")
-	}
-	if (length(algebra) && type != 'raw') {
-		stop("algebra only permitted for type='raw' data")
-	}
-	if (type == "cov") {
-		verifyCovarianceMatrix(observed)
-	}
-	if (type == "cor") {
-		verifyCorrelationMatrix(observed)
-	}
+  if (type != 'none') {
+    if (type != "raw" && is.na(numObs)) {
+      stop("Number of observations must be specified for non-raw data, i.e., add numObs=XXX to mxData()")
+    }
+    if (length(algebra) && type != 'raw') {
+      stop("algebra only permitted for type='raw' data")
+    }
+    if (type == "cov") {
+      verifyCovarianceMatrix(observed)
+    }
+    if (type == "cor") {
+      verifyCorrelationMatrix(observed)
+    }
+  }
 	if (!is.null(observedStats[['cov']])) {
 		verifyCovarianceMatrix(observedStats[['cov']], nameMatrix="covariance")
 	}
@@ -229,71 +241,76 @@ mxData <- function(observed, type, means = NA, numObs = NA, acov=NA, fullWeight=
 	}
 	thr <- observedStats[['thresholds']]
 	if (!is.null(thr)) {
-		verifyThresholdNames(thr, observed)
+		verifyThresholdNames(thr, colnames(observedStats[['cov']]))
 	}
-	lapply(dimnames(observed)[[2]], imxVerifyName, -1)
+  if (!is.null(observed)) {
+    lapply(dimnames(observed)[[2]], imxVerifyName, -1)
+  }
 	if(is.matrix(means)){meanNames <- colnames(means)} else {meanNames <- names(means)}
 	means <- as.matrix(means)
 	dim(means) <- c(1, length(means))
 	colnames(means) <- meanNames
 
-	if (length(primaryKey) > 1) {
-		stop("More than 1 primary key is not implemented yet")
-	} else if (!is.na(primaryKey)) {
-		if (type != "raw") {
-			stop(paste("Raw data is required when a primary key is provided"))
-		}
-		if (!(primaryKey %in% colnames(observed))) {
-			stop(paste("Primary key", omxQuotes(primaryKey),
-				   "must be provided in the observed data"))
-		}
-	}
+  if (!is.null(observed)) {
+    if (length(primaryKey) > 1) {
+      stop("More than 1 primary key is not implemented yet")
+    } else if (!is.na(primaryKey)) {
+      if (type != "raw") {
+        stop(paste("Raw data is required when a primary key is provided"))
+      }
+      if (!(primaryKey %in% colnames(observed))) {
+        stop(paste("Primary key", omxQuotes(primaryKey),
+                   "must be provided in the observed data"))
+      }
+    }
 
-	if (length(weight) > 1) {
-		stop("Only one weight column can be specified")
-	} else if (!is.na(weight)) {
-		if (type != "raw") {
-			stop(paste("Raw data is required when a weight column is provided"))
-		}
-		if (!(weight %in% colnames(observed))) {
-			stop(paste("Weight column", omxQuotes(weight),
-				   "must be provided in the observed data"))
-		}
-	}
-	if (length(frequency) > 1) {
-		stop("Only one frequency column can be specified")
-	} else if (!is.na(frequency)) {
-		if (type != "raw") {
-			stop(paste("Raw data is required when a frequency column is provided"))
-		}
-		if (!(frequency %in% colnames(observed))) {
-			stop(paste("Frequency column", omxQuotes(frequency),
-				   "must be provided in the observed data"))
-		}
-	}
-	if (type == "raw") {
-		if (!is.na(frequency)) {
-			obsCount <- sum(observed[,frequency])
-		} else {
-			obsCount <- nrow(observed)
-		}
+    if (length(weight) > 1) {
+      stop("Only one weight column can be specified")
+    } else if (!is.na(weight)) {
+      if (type != "raw") {
+        stop(paste("Raw data is required when a weight column is provided"))
+      }
+      if (!(weight %in% colnames(observed))) {
+        stop(paste("Weight column", omxQuotes(weight),
+                   "must be provided in the observed data"))
+      }
+    }
+    if (length(frequency) > 1) {
+      stop("Only one frequency column can be specified")
+    } else if (!is.na(frequency)) {
+      if (type != "raw") {
+        stop(paste("Raw data is required when a frequency column is provided"))
+      }
+      if (!(frequency %in% colnames(observed))) {
+        stop(paste("Frequency column", omxQuotes(frequency),
+                   "must be provided in the observed data"))
+      }
+    }
+    if (type == "raw") {
+      if (!is.na(frequency)) {
+        obsCount <- sum(observed[,frequency])
+      } else {
+        obsCount <- nrow(observed)
+      }
 
-		if (missing(numObs)) {
-			numObs <- obsCount
-		} else {
-			if (numObs != obsCount) {
-				# stop?
-				warning(paste("numObs of", numObs, "does not match the number of observations",
-					      "found in the observed data", obsCount,
-					      "(maybe specify weight='frequency' instead of numObs =",numObs,")"))
-			}
-		}
-	}
+      if (missing(numObs)) {
+        numObs <- obsCount
+      } else {
+        if (numObs != obsCount) {
+          # stop?
+          warning(paste("numObs of", numObs, "does not match the number of observations",
+                        "found in the observed data", obsCount,
+                        "(maybe specify weight='frequency' instead of numObs =",numObs,")"))
+        }
+      }
+    }
+  }
 
 	return(new("MxDataStatic", observed, means, type, as.numeric(numObs),
 		observedStats, sort, primaryKey, weight, frequency, as.integer(verbose),
 		as.logical(.parallel), as.logical(.noExoOptimize), minVariance,
-		as.character(algebra), as.logical(warnNPDacov), exoFree, naAction))
+		as.character(algebra), as.logical(warnNPDacov && warnNPDuseWeight), exoFree, naAction,
+    as.numeric(fitTolerance), as.numeric(gradientTolerance)))
 }
 
 setGeneric("preprocessDataForBackend", # DEPRECATED
@@ -405,7 +422,7 @@ setMethod("preprocessDataForBackend", signature("MxDataStatic"),
 			if(!is.null(data) && length(data@observedStats)) {
 				obsStats <- data@observedStats
 				if (!is.null(obsStats$thresholds)) {
-					verifyThresholdNames(obsStats$thresholds, data@observed, model@name)
+					verifyThresholdNames(obsStats$thresholds, data, model@name)
 				}
 			}
 			data
@@ -520,7 +537,7 @@ verifyCovarianceMatrix <- function(covMatrix, nameMatrix="observed", strictPD=TR
 	if (any(is.na(covMatrix))) {
 		msg <- paste("The", nameMatrix, "covariance matrix",
 			"contains NA values. Perhaps ensure you are excluding NAs in your cov() statement?")
-		stop(msg, call. = FALSE)	
+		stop(msg, call. = FALSE)
 	}
 	verifySymmetric(covMatrix, nameMatrix)
 	if (is.data.frame(covMatrix)) covMatrix <- as.matrix(covMatrix)
@@ -569,7 +586,7 @@ verifyCorrelationMatrix <- function(corMatrix) {
 	if (any(is.na(corMatrix))) {
 		msg <- paste("The observed correlation matrix",
 			"contains NA values")
-		stop(msg, call. = FALSE)	
+		stop(msg, call. = FALSE)
 	}
 	verifySymmetric(corMatrix)
 	if (is.data.frame(corMatrix)) corMatrix <- as.matrix(corMatrix)
@@ -587,7 +604,7 @@ verifyCorrelationMatrix <- function(corMatrix) {
 			msg <- paste("The observed correlation matrix",
 				"is not 1's along the diagonal")
 		}
-		stop(msg, call. = FALSE)	
+		stop(msg, call. = FALSE)
 	}
 }
 
@@ -613,13 +630,13 @@ displayMxData <- function(object) {
 		cat("primary key :", omxQuotes(object@primaryKey), '\n')
 	}
 	cat("numObs :", omxQuotes(object@numObs), '\n')
-	cat("observed : \n") 
+	cat("observed : \n")
 	print(object@observed)
 	if (length(object@means) == 1 && is.na(object@means)) {
 		cat("means : NA \n")
 	} else {
-		cat("means : \n") 
-		print(object@means)		
+		cat("means : \n")
+		print(object@means)
 	}
 	if (.hasSlot(object, 'observedStats') && length(object@observedStats)) {
 		cat("observedStats : \n")
@@ -629,11 +646,11 @@ displayMxData <- function(object) {
 }
 
 setMethod("print", "MxDataStatic", function(x,...) {
-	displayMxData(x) 
+	displayMxData(x)
 })
 
 setMethod("show", "MxDataStatic", function(object) {
-	displayMxData(object) 
+	displayMxData(object)
 })
 
 displayMxDataDynamic <- function(object) {
@@ -645,9 +662,9 @@ displayMxDataDynamic <- function(object) {
 }
 
 setMethod("print", "MxDataDynamic", function(x,...) {
-	displayMxDataDynamic(x) 
+	displayMxDataDynamic(x)
 })
 
 setMethod("show", "MxDataDynamic", function(object) {
-	displayMxDataDynamic(object) 
+	displayMxDataDynamic(object)
 })
