@@ -1,5 +1,5 @@
 #
-#   Copyright 2007-2019 by the individuals mentioned in the source code history
+#   Copyright 2007-2021 by the individuals mentioned in the source code history
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
 #   you may not use this file except in compliance with the License.
@@ -18,17 +18,21 @@ setClass(Class = "BaseExpectationNormal",
          representation = representation(
            expectedCovariance = "MxOptionalCharOrNumber",
            expectedMean = "MxOptionalCharOrNumber",
+           expectedFullCovariance = "MxOptionalCharOrNumber",
+           expectedFullMean = "MxOptionalCharOrNumber",
            thresholds = "MxCharOrNumber",
            threshnames = "character",
            discrete = "MxCharOrNumber",
            discreteSpec = "MxOptionalMatrix",
-           .discreteCheckCount = "logical"))
+           .discreteCheckCount = "logical",
+           selectionPlan = "MxOptionalDataFrame",
+           selectionVector = "MxCharOrNumber"))
 
 setMethod("qualifyNames", signature("BaseExpectationNormal"),
           function(.Object, modelname, namespace) {
             for (sl in c(paste0('expected', c('Covariance','Mean')),
-                       'thresholds', 'discrete')) {
-              if (is.null(slot(.Object, sl))) next
+                       'thresholds', 'discrete', 'selectionVector')) {
+              if (!.hasSlot(.Object, sl) || is.null(slot(.Object, sl))) next
               slot(.Object, sl) <- imxConvertIdentifier(slot(.Object, sl), modelname, namespace)
             }
             .Object
@@ -36,7 +40,7 @@ setMethod("qualifyNames", signature("BaseExpectationNormal"),
 
 setMethod("genericExpDependencies", signature("BaseExpectationNormal"),
           function(.Object, dependencies) {
-            sources <- c(.Object@discrete, .Object@thresholds)
+            sources <- c(.Object$discrete, .Object@thresholds, .Object$selectionVector)
             sources <- sources[!is.na(sources)]
             dependencies <- imxAddDependency(sources, .Object@name, dependencies)
             return(dependencies)
@@ -45,8 +49,8 @@ setMethod("genericExpDependencies", signature("BaseExpectationNormal"),
 setMethod("genericExpRename", signature("BaseExpectationNormal"),
           function(.Object, oldname, newname) {
             for (sl in c(paste0('expected', c('Covariance','Mean')),
-                         'thresholds', 'discrete')) {
-              if (is.null(slot(.Object, sl))) next
+                         'thresholds', 'discrete', 'selectionVector')) {
+              if (!.hasSlot(.Object, sl) || is.null(slot(.Object, sl))) next
               slot(.Object,sl) <- renameReference(slot(.Object,sl), oldname, newname)
             }
             .Object
@@ -96,6 +100,20 @@ setMethod("genericExpFunConvert", signature("BaseExpectationNormal"),
               }
             }
 
+            if (.hasSlot(.Object, 'selectionVector')) {
+              svname <- .Object$selectionVector
+              if (!is.na(svname)) {
+                sv <- flatModel[[svname]]
+                if (ncol(sv) != 1) stop("selectionVector must be a column vector")
+                svSpec <- .Object@selectionPlan
+                if (nrow(sv) != nrow(svSpec)) {
+                  msg <- paste('selectionPlan must have the same number of rows',
+                               'as the selectionVector')
+                  stop(msg, call.=FALSE)
+                }
+              }
+            }
+
             .Object
           })
 
@@ -103,7 +121,8 @@ setMethod("genericNameToNumber", signature("BaseExpectationNormal"),
 	  function(.Object, flatModel, model) {
 		  name <- .Object@name
       for (sl in c(paste0('expected', c('Covariance','Mean')),
-                   'thresholds', 'discrete')) {
+                   'data', 'thresholds', 'discrete', 'selectionVector')) {
+        if (!.hasSlot(.Object, sl)) next
         slot(.Object,sl) <- imxLocateIndex(flatModel, slot(.Object,sl), name)
       }
 		  .Object
@@ -238,7 +257,8 @@ setClass(Class = "MxExpectationNormal",
 
 setMethod("initialize", "MxExpectationNormal",
 	function(.Object, covariance, means, dims, thresholds, threshnames, discrete,
-           discreteSpec, data = as.integer(NA), name = 'expectation') {
+           discreteSpec, selectionVector, selectionPlan,
+           data = as.integer(NA), name = 'expectation') {
 		.Object@name <- name
 		.Object@covariance <- covariance
 		.Object@means <- means
@@ -249,6 +269,8 @@ setMethod("initialize", "MxExpectationNormal",
 		.Object@discrete <- discrete
 		.Object@discreteSpec <- discreteSpec
     .Object@.discreteCheckCount <- TRUE
+    .Object@selectionVector <- selectionVector
+    .Object@selectionPlan <- selectionPlan
 		.Object@ExpCov <- matrix()
 		.Object@ExpMean <- matrix()
 		return(.Object)
@@ -377,7 +399,8 @@ setMethod("genericGetExpectedVector", signature("BaseExpectationNormal"),
 
 setMethod("genericGetExpectedStandVector", signature("BaseExpectationNormal"),
 	function(.Object, model, defvar.row=1, subname=model@name) {
-		ret <- genericGetExpected(.Object, model, c('covariance', 'means', 'thresholds'), defvar.row, subname)
+		ret <- suppressWarnings(genericGetExpected(
+      .Object, model, c('covariance', 'means', 'thresholds', 'slope'), defvar.row, subname))
 		cov <- ret[['covariance']]
 		mns <- ret[['means']]
 		if (is.null(mns)) stop("mns is null")
@@ -387,11 +410,11 @@ setMethod("genericGetExpectedStandVector", signature("BaseExpectationNormal"),
 			thrNames <- outer(paste0('thr', 1:nrow(thr)), 1:ncol(thr), paste, sep='_')
 			dth <- getThresholdMask(model, colnames(thr), subname)
 		}
-		v <- .standardizeCovMeansThresholds(cov, mns, thr, dth, vector=TRUE)
+		v <- .standardizeCovMeansThresholds(cov, mns, thr, ret[['slope']], dth, vector=TRUE)
 		return(v)
 })
 
-.standardizeCovMeansThresholds <- function(cov, means, thresholds, dth, vector=FALSE){
+.standardizeCovMeansThresholds <- function(cov, means, thresholds, slope, dth, vector=FALSE){
   mnames <- colnames(cov)
   if (length(mnames) == 0) {
     stop("I give up. Have no idea how to standardize this expectation.\nYour covariance matrix must have dimnames.")
@@ -405,7 +428,7 @@ setMethod("genericGetExpectedStandVector", signature("BaseExpectationNormal"),
     cov <- .ordinalCov2Cor(cov, ordInd)
   }
 	if(!vector){
-		return(list(cov=cov, means=means, thresholds=thresholds))
+		return(list(cov=cov, means=means, thresholds=thresholds, slope=slope))
 	} else {
 		v <- c()
 		vn <- c()
@@ -422,6 +445,12 @@ setMethod("genericGetExpectedStandVector", signature("BaseExpectationNormal"),
 		    }
 		  }
 		}
+    if (length(slope)) {
+      v <- c(v, slope)
+      sname <- apply(expand.grid(row=rownames(slope),
+                                 col=colnames(slope))[,c(2,1)], 1, paste, collapse='_')
+      vn <- c(vn, sname)
+    }
 		for (vx in 1:length(mnames)) {
 			if (any(vx == ordInd)) next
 			v <- c(v, cov[vx,vx])
@@ -535,6 +564,7 @@ mxCheckIdentification <- function(model, details=TRUE){
 	theParams <- omxGetParameters(model)
 	jac <- omxManifestModelByParameterJacobian(model)
 	if(imxHasConstraint(model)){
+    # Better if there was a separate compute step to estimate the constraintJacobian
 		tmpModel <- mxModel(model, mxComputeSequence(list(mxComputeNumericDeriv(hessian=FALSE), mxComputeReportDeriv())))
 		tmpModel <- mxRun(tmpModel, silent=TRUE)
 		cjac <- tmpModel$output$constraintJacobian
@@ -580,15 +610,15 @@ mxCheckIdentification <- function(model, details=TRUE){
 
 setGeneric("genericGenerateData",
 	function(.Object, model, nrows, subname, empirical, returnModel, use.miss,
-		   .backend, nrowsProportion) {
+		   .backend, nrowsProportion, silent) {
 	return(standardGeneric("genericGenerateData"))
 })
 
 setMethod("genericGenerateData", signature("MxExpectationNormal"),
 	  function(.Object, model, nrows, subname, empirical, returnModel, use.miss,
-		   .backend, nrowsProportion) {
+		   .backend, nrowsProportion, silent) {
 	    return(generateNormalData(model, nrows, subname, empirical, returnModel,
-				      use.miss, .backend, nrowsProportion))
+				      use.miss, .backend, nrowsProportion, silent))
 	  })
 
 .rmvnorm <- function(nrow, mean, sigma, empirical) {
@@ -617,13 +647,16 @@ calcNumRows <- function(nrows, nrowsProportion, origRows, subname) {
 }
 
 generateNormalData <- function(model, nrows, subname, empirical, returnModel, use.miss,
-			       .backend, nrowsProportion) {
+			       .backend, nrowsProportion, silent) {
   origData <- findDataForSubmodel(model, subname)
   origRows <- if (!is.null(origData)) { nrowMxData(origData) } else { NULL }
   nrows <- calcNumRows(nrows, nrowsProportion, origRows, subname)
 
 	# Check for definition variables
 	if(imxHasDefinitionVariable(model[[subname]])){
+    if (is.null(origData)) {
+      stop("Definition variable(s) found, but no data is available")
+    }
 		if (origData$type != 'raw') {
 			stop(paste("Definition variable(s) found, but original data is type",
 				omxQuotes(origData$type)))
@@ -637,12 +670,17 @@ generateNormalData <- function(model, nrows, subname, empirical, returnModel, us
 		data <- matrix(NA, nrow=nrows, ncol=ncol(theCov))
 		colnames(data) <- colnames(theCov)
 		data <- as.data.frame(data)
+    prevProgressLen <- 0L
 		for(i in 1:nrows){
 			theMeans <- imxGetExpectationComponent(model, "means", defvar.row=i, subname=subname)
 			theCov <- imxGetExpectationComponent(model, "covariance", defvar.row=i, subname=subname)
 			theThresh <- imxGetExpectationComponent(model, "thresholds", defvar.row=i, subname=subname)
 			data[i,] <- .rmvnorm(1, theMeans, theCov, empirical)
+      info <- paste0(subname,' ', i, '/', nrows)
+      if (!silent) imxReportProgress(info, prevProgressLen)
+      prevProgressLen <- nchar(info)
 		}
+    if (!silent) imxReportProgress('', prevProgressLen)
 		data <- ordinalizeDataHelper(data, theThresh, origData)
 		if (!is.null(origData)) {
 			for (dcol in setdiff(colnames(origData), colnames(data))) {
@@ -725,31 +763,27 @@ generateRelationalData <- function(model, returnModel, .backend, subname, empiri
 		modelE <- mxRun(modelE, silent=TRUE)
 		simData <- modelE$compute$output
 
-		datalist <- modelE@runstate$datalist
-		for (dName in names(datalist)) {
-			if (is.null(simData[[dName]])) {
-				simData[[dName]] <- datalist[[dName]]$observed
-			} else {
-				orig <- datalist[[dName]]$observed
-				toCopy <- setdiff(colnames(orig), colnames(simData[[dName]]))
-				for (col in toCopy) {
-					simData[[dName]][[col]] <- orig[[col]]
-				}
-			}
-		}
-
 		names(simData) <- substr(names(simData), 1, nchar(names(simData))-5) #strip .data
 
+    for (modelName in names(simData)) {
+      if (modelName == model$name) {
+        dat <- model@data@observed
+      } else {
+        dat <- model[[modelName]]@data@observed
+      }
+      newDat <- simData[[modelName]]
+      toCopy <- setdiff(colnames(dat), colnames(newDat))
+      for (col in toCopy) newDat[[col]] <- dat[[col]]
+      simData[[modelName]] <- newDat
+      if (modelName == model$name) {
+        model@data@observed <- newDat
+      } else {
+        model[[modelName]]@data@observed <- newDat
+      }
+    }
 		if (!returnModel) {
 			return(simData)
 		} else {
-			for (modelName in names(simData)) {
-				if (modelName == model$name) {
-					model@data@observed <- simData[[modelName]]
-				} else {
-					model[[modelName]]@data@observed <- simData[[modelName]]
-				}
-			}
 			return(model)
 		}
 	}
@@ -761,12 +795,14 @@ generateRelationalData <- function(model, returnModel, .backend, subname, empiri
 	))
 
 	modelE <- mxModel(model, plan)
+  modelE$expectation$.maxDebugGroups <- 100L
 	modelE$expectation$.rampartCycleLimit <- 0L
 	modelE <- mxRun(modelE, silent=TRUE)
 	dataEnv <- new.env()
-	for (dName in names(modelE@runstate$datalist)) {
+  datasets <- collectDatasets(modelE, imxGenerateNamespace(model))
+	for (dName in names(datasets)) {
 		modelName <- substr(dName, 1, nchar(dName)-5)  # remove .data
-		assign(modelName, modelE@runstate$datalist[[ dName ]]@observed, envir=dataEnv)
+		assign(modelName, datasets[[ dName ]]@observed, envir=dataEnv)
 	}
 	ed <- modelE$expectation$debug
 	layout <- ed$layout
@@ -844,7 +880,7 @@ extractObservedData <- function(model) {
 }
 
 mxGenerateData <- function(model, nrows=NULL, returnModel=FALSE, use.miss = TRUE,
-			   ..., .backend=TRUE, subname=NULL, empirical=FALSE, nrowsProportion=NULL) {
+			   ..., .backend=TRUE, subname=NULL, empirical=FALSE, nrowsProportion=NULL, silent=FALSE) {
 	prohibitDotdotdot(list(...))
 	if (!is.null(nrows) && !is.null(nrowsProportion)) {
 	  stop("You cannot specify both nrows and nrowsProportion")
@@ -887,7 +923,7 @@ mxGenerateData <- function(model, nrows=NULL, returnModel=FALSE, use.miss = TRUE
 		}
 	}
 	genericGenerateData(model[[subname]]$expectation, model, nrows, subname, empirical,
-			    returnModel, use.miss, .backend, nrowsProportion)
+			    returnModel, use.miss, .backend, nrowsProportion, silent)
 }
 
 verifyExpectedObservedNames <- function(data, covName, flatModel, modelname, objectiveName) {
@@ -936,7 +972,6 @@ setMethod("genericExpFunConvert", "MxExpectationNormal",
 
 		mxDataObject <- flatModel@datasets[[.Object@data]]
 		dataName <- .Object@data
-		.Object@data <- imxLocateIndex(flatModel, .Object@data, name)
 		threshName <- .Object@thresholds
 		covName <- .Object@covariance
 		covariance <- flatModel[[covName]]
@@ -946,7 +981,7 @@ setMethod("genericExpFunConvert", "MxExpectationNormal",
 
 		if (inherits(mxDataObject, "MxDataDynamic")) return(.Object)
 
-		if (mxDataObject@type != "raw") {
+		if (!(mxDataObject@type %in% c("none","raw"))) {
 			verifyExpectedObservedNames(mxDataObject@observed, covName, flatModel, modelname, "Normal")
 			verifyMeans(meansName, mxDataObject, flatModel, modelname)
 		}
@@ -960,9 +995,32 @@ setMethod("genericExpFunConvert", "MxExpectationNormal",
 		if (single.na(.Object@dims)) {
 			.Object@dims <- covNames
 		}
+
+    if (.hasSlot(.Object, 'selectionVector')) {
+      .Object@selectionPlan <- prepSelectionPlan(.Object@selectionPlan, covNames)
+    }
+
 		return(.Object)
 })
 
+prepSelectionPlan <- function(selPlan, covNames) {
+  if (is.null(selPlan)) return(NULL)
+
+  ind <- match(selPlan$from, covNames)
+  if (any(is.na(ind))) {
+    stop(paste("Can't find selectionPlan from variables",
+               omxQuotes(selPlan$from[is.na(ind)])))
+  }
+  selPlan$from <- ind - 1L
+  ind <- match(selPlan$to, covNames)
+  if (any(is.na(ind))) {
+    stop(paste("Can't find selectionPlan to variables",
+               omxQuotes(selPlan$to[is.na(ind)])))
+  }
+  selPlan$to <- ind - 1L
+  selPlan
+}
+
 verifyMvnNames <- function(covName, meansName, type, flatModel, modelname, expectationName) {
 	if (is.na(meansName)) {
 		means <- NA
@@ -1012,6 +1070,7 @@ verifyMvnNames <- function(covName, meansName, type, flatModel, modelname, expec
 }
 
 verifyObservedNames <- function(data, means, type, flatModel, modelname, expectationName) {
+  if (is.null(data)) return()
 	dataNames <- dimnames(data)
 	if(is.null(dataNames)) {
 		msg <- paste("The observed data associated with the",
@@ -1051,7 +1110,16 @@ verifyObservedNames <- function(data, means, type, flatModel, modelname, expecta
 generateDataColumns <- function(flatModel, covNames, dataName) {
 	retval <- c()
 	if (length(covNames) == 0) return(retval)
-	dataColumnNames <- colnames(flatModel@datasets[[dataName]]@observed)
+  mxd <- flatModel@datasets[[dataName]]
+  if (!is.null(mxd@observed)) {
+    dataColumnNames <- colnames(mxd@observed)
+  } else if (!is.null(mxd$observedStats[['cov']])) {
+    dataColumnNames <- colnames(mxd$observedStats[['cov']])
+  } else {
+    msg <- paste(omxQuotes(dataName), "is type='raw' but",
+                 "raw observed data seems to be unavailable")
+    stop(msg, call.=FALSE)
+  }
 	retval <- match(covNames, dataColumnNames)
 	if (any(is.na(retval))) {
 		msg <- paste("The column name(s)", omxQuotes(covNames[is.na(retval)]),
@@ -1175,13 +1243,13 @@ updateExpectationDimnames <- function(flatExpectation, flatModel,
 	}
 
 	if (!is.null(dimnames(means)) && !single.na(dims) &&
-		!identical(dimnames(means), list(NULL, dims))) {
+		!identical(colnames(means), dims)) {
 		modelname <- getModelName(flatExpectation)
 		msg <- paste("The expected means matrix associated",
 			"with the expectation function in model",
-			omxQuotes(modelname), "contains dimnames: ",
-            paste(toString(dimnames(means)), ".", sep = ""),
-			"The expectation function has specified dimnames:",
+			omxQuotes(modelname), "contains colnames: ",
+            paste(toString(colnames(means)), ".", sep = ""),
+			"The expectation function has specified colnames:",
 			paste(toString(dims), ".", sep =""))
 		stop(msg, call.=FALSE)
 	}
@@ -1261,8 +1329,10 @@ mxExpectationNormal <-
 		stop("NA values are not allowed for 'dimnames' vector")
 	}
 	threshnames <- checkThreshnames(threshnames)
+  selectionVector <- NA_character_
+  selectionPlan <- NULL
 	return(new("MxExpectationNormal", covariance, means, dimnames, thresholds, threshnames,
-             discrete, discreteSpec))
+             discrete, discreteSpec, selectionVector, selectionPlan))
 }
 
 displayMxExpectationNormal <- function(expectation) {
