@@ -1,5 +1,5 @@
-#ifndef _LOAD_DATA_API_H_
-#define _LOAD_DATA_API_H_
+#ifndef u_LOAD_DATA_API_H_
+#define u_LOAD_DATA_API_H_
 
 #include <memory>
 
@@ -68,11 +68,11 @@ protected:
 
 public:
 	const std::vector< int > &getColumns() { return columns; }
-	void commonInit(SEXP rObj, const char *_name,
-                  const char *_dataName, int rows,
-                  std::vector<ColumnData> &_rawCols,
-                  ColMapType &_rawColMap,
-                  std::vector< std::string > &_checkpointValues,
+	void commonInit(SEXP rObj, const char *u_name,
+                  const char *u_dataName, int rows,
+                  std::vector<ColumnData> &u_rawCols,
+                  ColMapType &u_rawColMap,
+                  std::vector< std::string > &u_checkpointValues,
                   bool useOriginalData);
 	virtual int getNumVariants() { return 0; }
 	bool wantCheckpoint() const { return checkpoint; }
@@ -106,17 +106,18 @@ public:
 	}
 	virtual std::unique_ptr<LoadDataProviderBase2> clone()=0;
 	virtual ~LoadDataProviderBase2() {
-		int stripes = stripeData.size() / columns.size();
-		for (int sx=0; sx < stripes; ++sx) {
-			for (int cx=0; cx < int(columns.size()); ++cx) {
-				int dx = sx * columns.size() + cx;
-				if (colTypes[cx] == COLUMNDATA_NUMERIC) {
-					delete [] stripeData[dx].realData;
-				} else {
-					delete [] stripeData[dx].intData;
-				}
-			}
-		}
+    if (!columns.size()) return;
+    int stripes = stripeData.size() / columns.size();
+    for (int sx=0; sx < stripes; ++sx) {
+      for (int cx=0; cx < int(columns.size()); ++cx) {
+        int dx = sx * columns.size() + cx;
+        if (colTypes[cx] == COLUMNDATA_NUMERIC) {
+          delete [] stripeData[dx].realData;
+        } else {
+          delete [] stripeData[dx].intData;
+        }
+      }
+    }
 		stripeData.clear();
 	}
 };
@@ -124,7 +125,7 @@ public:
 template <typename Derived>
 class LoadDataProvider : public LoadDataProviderBase2 {
 public:
-	virtual std::unique_ptr<LoadDataProviderBase2> clone() {
+	virtual std::unique_ptr<LoadDataProviderBase2> clone() override {
 		return std::unique_ptr<LoadDataProviderBase2>(new Derived());
 	}
 };
@@ -133,6 +134,6 @@ public:
 //#define OPENMX_LOAD_DATA_API_VERSION 0.3091921037994325
 #define OPENMX_LOAD_DATA_API_VERSION 0.5240939254872501 // this is a random number
 
-typedef void (*AddLoadDataProviderType)(double version, int ldpbSz, LoadDataProviderBase2 *ldp);
+typedef void (*AddLoadDataProviderType)(double version, unsigned int ldpbSz, LoadDataProviderBase2 *ldp);
 
 #endif
