@@ -1,5 +1,5 @@
-#ifndef _RAMINTERNAL_H_
-#define _RAMINTERNAL_H_
+#ifndef u_RAMINTERNAL_H_
+#define u_RAMINTERNAL_H_
 
 #include <Eigen/Core>
 #include <Eigen/SparseCore>
@@ -18,8 +18,8 @@ struct coeffLoc {
 	int off;
 	int r, c;
 
-coeffLoc(int _off, int _r, int _c) :
-	off(_off), r(_r), c(_c) {}
+coeffLoc(int u_off, int u_r, int u_c) :
+	off(u_off), r(u_r), c(u_c) {}
 };
 
 class omxRAMExpectation;
@@ -51,7 +51,7 @@ namespace RelationalRAMExpectation {
 		omxExpectation *model;  // read-only
 	public:
 		int row;                 // to load definition variables (never the key)
-		struct independentGroup *ig;
+		class independentGroup *ig;
 		int igIndex;
 		int nextMean;
 
@@ -121,11 +121,11 @@ namespace RelationalRAMExpectation {
 		struct MpcIO : PathCalcIO {
 			independentGroup &par;
 			int clumpSize;
-			MpcIO(independentGroup &_par) : par(_par), clumpSize(_par.clumpSize) {}
-			virtual void recompute(FitContext *fc);
-			virtual unsigned getVersion(FitContext *fc);
-			virtual void refresh(FitContext *fc);
-			virtual PathCalcIO *clone()
+			MpcIO(independentGroup &u_par) : par(u_par), clumpSize(u_par.clumpSize) {}
+			virtual void recompute(FitContext *fc) override;
+			virtual unsigned getVersion(FitContext *fc) override;
+			virtual void refresh(FitContext *fc) override;
+			virtual PathCalcIO *clone() override
 			{ return new MpcIO(par); }
 		};
 
@@ -133,32 +133,32 @@ namespace RelationalRAMExpectation {
 			independentGroup &par;
 			int clumpSize;
 			bool useRampart;
-			ApcIO(independentGroup &_par) : par(_par), clumpSize(_par.clumpSize), useRampart(true) {}
-			virtual void recompute(FitContext *fc);
-			virtual unsigned getVersion(FitContext *fc);
+			ApcIO(independentGroup &u_par) : par(u_par), clumpSize(u_par.clumpSize), useRampart(true) {}
+			virtual void recompute(FitContext *fc) override;
+			virtual unsigned getVersion(FitContext *fc) override;
 			template <typename T>
-			void _refresh(FitContext *fc, T &mat, double sign);
-			virtual void refreshA(FitContext *fc, double sign)
-			{ _refresh(fc, full, sign); }
-			virtual void refreshSparse1(FitContext *fc, double sign)
-			{ _refresh(fc, sparse, sign); }
-			virtual PathCalcIO *clone()
+			void u_refresh(FitContext *fc, T &mat, double sign);
+			virtual void refreshA(FitContext *fc, double sign) override
+			{ u_refresh(fc, full, sign); }
+			virtual void refreshSparse1(FitContext *fc, double sign) override
+			{ u_refresh(fc, sparse, sign); }
+			virtual PathCalcIO *clone() override
 			{ return new ApcIO(par); }
 		};
 
 		struct SpcIO : PathCalcIO {
 			independentGroup &par;
 			int clumpSize;
-			SpcIO(independentGroup &_par) : par(_par), clumpSize(_par.clumpSize) {}
-			virtual void recompute(FitContext *fc);
-			virtual unsigned getVersion(FitContext *fc);
+			SpcIO(independentGroup &u_par) : par(u_par), clumpSize(u_par.clumpSize) {}
+			virtual void recompute(FitContext *fc) override;
+			virtual unsigned getVersion(FitContext *fc) override;
 			template <typename T>
-			void _refresh(FitContext *fc, T &mat);
-			virtual void refresh(FitContext *fc)
-			{ _refresh(fc, full); }
-			virtual void refreshSparse1(FitContext *fc, double sign)
-			{ _refresh(fc, sparse); }
-			virtual PathCalcIO *clone()
+			void u_refresh(FitContext *fc, T &mat);
+			virtual void refresh(FitContext *fc) override
+			{ u_refresh(fc, full); }
+			virtual void refreshSparse1(FitContext *fc, double sign) override
+			{ u_refresh(fc, sparse); }
+			virtual PathCalcIO *clone() override
 			{ return new SpcIO(par); }
 		};
 
@@ -188,7 +188,7 @@ namespace RelationalRAMExpectation {
 		PathCalc pcalc;
 		double                           fit;  // most recent fit for debugging
 
-		independentGroup(class state *_st, int size, int _clumpSize);
+		independentGroup(class state *u_st, int size, int u_clumpSize);
 		independentGroup(independentGroup *ig);
 		int numLooseClumps() {
 			independentGroup &par = getParent();
@@ -225,8 +225,8 @@ namespace RelationalRAMExpectation {
 	public:
 		bool isChild() const { return this != parent; }
 		typedef std::vector< std::set<int> > SubgraphType;
-		struct omxExpectation *homeEx;
-		std::set<struct omxExpectation *> allEx;
+		class omxExpectation *homeEx;
+		std::set<class omxExpectation *> allEx;
 		bool hasProductNodes;
 		typedef std::map< std::pair<omxData*,int>, int, RowToLayoutMapCompare> RowToLayoutMapType;
 		RowToLayoutMapType               rowToLayoutMap;
@@ -273,18 +273,19 @@ namespace RelationalRAMExpectation {
 		void simulate(FitContext *fc, MxRList &out);
 		int getOptimizeMean();
 		void optimizeModelRotation();
+    int numObservedStats();
 	};
 };
 
-class omxRAMExpectation : public omxExpectation {
-	typedef omxExpectation super;
+class omxRAMExpectation : public MVNExpectation {
+	typedef MVNExpectation super;
 	Eigen::VectorXi dataCols;  // composition of F permutation and expectation->dataColumns
 	std::vector<const char *> dataColNames;
 	std::vector< omxThresholdColumn > thresholds;
 	std::vector<int> exoDataColumns; // index into omxData
-	Eigen::VectorXd exoPredMean;
 	bool hasProductNodes;
   bool studiedF;
+  bool openBox;  // can the user access the expectation during optimization?
 	int numExoPred;
 	std::vector<int> exoDataColIndex;
   void addSlopeMatrix();
@@ -292,10 +293,10 @@ class omxRAMExpectation : public omxExpectation {
 	struct MpcIO : PathCalcIO {
 		omxMatrix *M0;
 		MpcIO() {}
-		virtual void recompute(FitContext *fc);
-		virtual unsigned getVersion(FitContext *fc);
-		virtual void refresh(FitContext *fc);
-		virtual PathCalcIO *clone()
+		virtual void recompute(FitContext *fc) override;
+		virtual unsigned getVersion(FitContext *fc) override;
+		virtual void refresh(FitContext *fc) override;
+		virtual PathCalcIO *clone() override
 		{
 			auto *mio = new MpcIO;
 			mio->M0 = M0;
@@ -306,16 +307,16 @@ class omxRAMExpectation : public omxExpectation {
 	struct ApcIO : PathCalcIO {
 		omxMatrix *A0;
 		std::vector<coeffLoc> &vec;
-		ApcIO(std::vector<coeffLoc> &_vec) : vec(_vec) {}
-		virtual void recompute(FitContext *fc);
-		virtual unsigned getVersion(FitContext *fc);
+		ApcIO(std::vector<coeffLoc> &u_vec) : vec(u_vec) {}
+		virtual void recompute(FitContext *fc) override;
+		virtual unsigned getVersion(FitContext *fc) override;
 		template <typename T>
-		void _refresh(FitContext *fc, T &mat, double sign);
-		virtual void refreshA(FitContext *fc,double sign)
-		{ _refresh(fc, full, sign); }
-		virtual void refreshSparse1(FitContext *fc, double sign)
-		{ _refresh(fc, sparse, sign); }
-		virtual PathCalcIO *clone()
+		void u_refresh(FitContext *fc, T &mat, double sign);
+		virtual void refreshA(FitContext *fc,double sign) override
+		{ u_refresh(fc, full, sign); }
+		virtual void refreshSparse1(FitContext *fc, double sign) override
+		{ u_refresh(fc, sparse, sign); }
+		virtual PathCalcIO *clone() override
 		{
 			auto *aio = new ApcIO(vec);
 			aio->A0 = A0;
@@ -326,16 +327,16 @@ class omxRAMExpectation : public omxExpectation {
 	struct SpcIO : PathCalcIO {
 		omxMatrix *S0;
 		std::vector<coeffLoc> &vec;
-		SpcIO(std::vector<coeffLoc> &_vec) : vec(_vec) {}
-		virtual void recompute(FitContext *fc);
-		virtual unsigned getVersion(FitContext *fc);
+		SpcIO(std::vector<coeffLoc> &u_vec) : vec(u_vec) {}
+		virtual void recompute(FitContext *fc) override;
+		virtual unsigned getVersion(FitContext *fc) override;
 		template <typename T>
-		void _refresh(FitContext *fc, T &mat);
-		virtual void refresh(FitContext *fc)
-		{ _refresh(fc, full); }
-		virtual void refreshSparse1(FitContext *fc, double sign)
-		{ _refresh(fc, sparse); }
-		virtual PathCalcIO *clone()
+		void u_refresh(FitContext *fc, T &mat);
+		virtual void refresh(FitContext *fc) override
+		{ u_refresh(fc, full); }
+		virtual void refreshSparse1(FitContext *fc, double sign) override
+		{ u_refresh(fc, sparse); }
+		virtual PathCalcIO *clone() override
 		{
 			auto *sio = new SpcIO(vec);
 			sio->S0 = S0;
@@ -367,9 +368,11 @@ class omxRAMExpectation : public omxExpectation {
 	void CalculateRAMCovarianceAndMeans(FitContext *fc);
 	void analyzeDefVars(FitContext *fc);
 	void logDefVarsInfluence();
+  bool isOpenBox() const { return openBox; }
 
 	omxMatrix *cov, *means; // observed covariance and means
 	omxMatrixPtr covOwner, meanOwner;
+	omxMatrix *fullCov, *fullMean;
 	omxMatrix *slope;       // exogenous predictor slopes
 	omxMatrix *A, *S, *F, *M;
 
@@ -393,30 +396,31 @@ class omxRAMExpectation : public omxExpectation {
 	void studyF();
 	void studyExoPred();
 
-	virtual void init();
-  virtual void connectToData();
-	virtual void compute(FitContext *fc, const char *what, const char *how);
-	virtual omxMatrix *getComponent(const char*);
-	virtual void populateAttr(SEXP expectation);
-	virtual const std::vector<const char *> &getDataColumnNames() const {
+	virtual void init() override;
+	virtual void compute(FitContext *fc, const char *what, const char *how) override;
+	virtual omxMatrix *getComponent(const char*) override;
+	virtual void populateAttr(SEXP expectation) override;
+	virtual const std::vector<const char *> &getDataColumnNames() const override {
     if (studiedF) return dataColNames;
     else return super::getDataColumnNames();
   };
-	virtual const Eigen::Map<DataColumnIndexVector> getDataColumns() {
+	virtual const Eigen::Map<DataColumnIndexVector> getDataColumns() override {
     if (studiedF)
       return Eigen::Map<DataColumnIndexVector>(dataCols.data(), numDataColumns);
     else
       return super::getDataColumns();
 	}
-	virtual std::vector< omxThresholdColumn > &getThresholdInfo()
+	virtual std::vector< omxThresholdColumn > &getThresholdInfo() override
   {
     if (studiedF) return thresholds;
     else return super::getThresholdInfo();
   }
-	virtual void invalidateCache();
-	virtual void generateData(FitContext *fc, MxRList &out);
-	virtual void flatten(FitContext *fc);
-	virtual void getExogenousPredictors(std::vector<int> &out);
+	virtual void invalidateCache() override;
+	virtual void generateData(FitContext *fc, MxRList &out) override;
+	void flatten(FitContext *fc);
+	virtual void getExogenousPredictors(std::vector<int> &out) override;
+  virtual int numLatentVars() const override;
+  virtual int numObservedStats() override;
 };
 
 namespace RelationalRAMExpectation {
