@@ -1,5 +1,5 @@
 /*
- *  Copyright 2007-2019 by the individuals mentioned in the source code history
+ *  Copyright 2007-2021 by the individuals mentioned in the source code history
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -99,8 +99,7 @@ void omxExpectation::compute(FitContext *fc, const char *what, const char *how)
     auto ds = getDiscreteSpec();
     auto dc = getDataColumns();
     if (data->isRaw()) {
-      // This is a bit clumsy. We should do this before cloning read-only data for threads. TODO
-      bool firstTime = discreteCache.size() == 0 && !isClone();
+      bool firstTime = discreteCache.size() == 0 && isTopState();
       for(int dx = 0; dx < int(dc.size()); dx++) {
         omxThresholdColumn &col = allTh[dx];
         if (!col.isDiscrete) continue;
@@ -110,10 +109,8 @@ void omxExpectation::compute(FitContext *fc, const char *what, const char *how)
           continue;
         }
         ColumnData &cd = data->rawCol(col.dataColumn);
-        const auto range =
-          std::minmax_element(cd.i(), cd.i() + data->nrows());
-        int obsMaxCount = *range.second - cd.getMinValue();
-        //mxLog("infer num thresholds for %s is %d", data->columnName(col.dataColumn), obsMaxCount);
+        int obsMaxCount = cd.getNumThresholds();
+        //mxLog("num thresholds for %s is %d", data->columnName(col.dataColumn), obsMaxCount);
         if (std::isfinite(nt)) {
           if (nt < obsMaxCount) {
             mxThrow("%s: discrete column '%s' set to a maximum count of %d "
@@ -206,6 +203,7 @@ void omxExpectation::loadThresholds()
 
 	//omxPrint(thresholdsMat, "loadThr");
 
+  auto dcn = base::getDataColumnNames();
 	auto dc = base::getDataColumns();
 	thresholds.resize(dc.size());
 	for(int dx = 0; dx < int(dc.size()); dx++) {
@@ -218,7 +216,7 @@ void omxExpectation::loadThresholds()
 	for(int dx = 0; dx < int(dc.size()); dx++) {
 		omxThresholdColumn &col = thresholds[dx];
 		int index = col.dataColumn;
-		const char *colname = data->columnName(index);
+		const char *colname = data->isRaw()? data->columnName(index) : dcn[index];
 
 		if (thresholdsMat) {
 			int tc = thresholdsMat->lookupColumnByName(colname);
@@ -227,7 +225,7 @@ void omxExpectation::loadThresholds()
 				col.column = tc;
 				col.isDiscrete = false;
 				if (data->isRaw()) {
-					col.numThresholds = omxDataGetNumFactorLevels(data, index) - 1;
+					col.numThresholds = data->rawCol(index).getNumThresholds();
 				} else {
 					// See omxData
 				}
@@ -242,7 +240,7 @@ void omxExpectation::loadThresholds()
 				col.column = tc;
 				col.isDiscrete = true;
         double nt = ds(0,tc);
-        col.numThresholds = nt; // can be NA
+        col.numThresholds = cast_with_NA(nt);
 				numOrdinal++;
 			}
 		}
@@ -282,6 +280,7 @@ void omxExpectation::connectToData()
   setConnectedToData(true);
 
 	if (!strEQ(omxDataType(data), "raw")) return;
+  if (data->nrows() == 0 && data->hasSummaryStats()) return;
 
 	auto &allTh = getThresholdInfo();
 
@@ -323,10 +322,11 @@ void omxExpectation::populateNormalAttr(SEXP robj, MxRList &out)
 {
   if (!discreteMat && !thresholdsMat) return;
 
+  auto &dcn = getDataColumnNames();
+
   if (discreteMat) { // update discreteSpec
     auto &allTh = getThresholdInfo();
     auto ds = getDiscreteSpec();
-    auto &dcn = getDataColumnNames();
     CharacterVector cn(ds.cols());
     Eigen::MatrixXd newDS(ds.rows(), ds.cols());
 
@@ -351,7 +351,7 @@ void omxExpectation::populateNormalAttr(SEXP robj, MxRList &out)
     for (int xx = 0, cx=0; cx < int(allTh.size()); cx++) {
       auto &th = allTh[cx];
       if (th.numThresholds == 0) continue;
-      cn[xx] = data->columnName(th.dataColumn);
+      cn[xx] = data->isRaw()? data->columnName(th.dataColumn) : dcn[cx];
       xx += 1;
     }
 
@@ -453,11 +453,11 @@ omxNewIncompleteExpectation(SEXP rObj, int expNum, omxState* os)
 	return expect;
 }
 
-void omxExpectation::setConnectedToData(bool _to)
+void omxExpectation::setConnectedToData(bool u_to)
 {
-  //mxLog("%s: connectedToData=%d", name, _to);
-  if (_to && getConnectedToData()) mxThrow("omxExpectation::connectToData() called again");
-  _connectedToData = _to;
+  //mxLog("%s: connectedToData=%d", name, u_to);
+  if (u_to && getConnectedToData()) mxThrow("omxExpectation::connectToData() called again");
+  u_connectedToData = u_to;
 }
 
 void omxCompleteExpectation(omxExpectation *ox) {
@@ -492,7 +492,7 @@ void omxCompleteExpectation(omxExpectation *ox) {
 const std::vector<const char *> &omxExpectation::getDataColumnNames() const
 { return dataColumnNames; }
 
-const Eigen::Map<omxExpectation::DataColumnIndexVector> omxExpectation::getDataColumns()
+const Eigen::Map<omxExpectation::DataColumnIndexVector> omxExpectation::getDataColumns() // NOT const
 {
 	return Eigen::Map<DataColumnIndexVector>(dataColumnsPtr, numDataColumns);
 }
@@ -612,5 +612,41 @@ void omxExpectation::asVector1(FitContext *fc, int row, Eigen::Ref<Eigen::Vector
 										getThresholdInfo(), out);
 }
 
-bool omxExpectation::isClone() const
-{ return currentState->isClone(); }
+bool omxExpectation::isTopState() const
+{ return currentState->isTopState(); }
+
+int MVNExpectation::numObservedStats()
+{
+  if (strEQ(data->getType(), "raw")) {
+    if (data->hasSummaryStats()) return numSummaryStats();
+
+    auto &dc = getDataColumns();
+    // More accurate to use sufficient statistics for each missingness pattern? TODO
+    int stats = 0;
+    for (int cx=0; cx < int(dc.size()); ++cx) {
+      stats += data->countObs(dc[cx]);
+    }
+
+    auto &allTh = getThresholdInfo();
+    for (auto &col : allTh) stats += col.numThresholds;
+
+    return stats;
+
+  } else if (strEQ(data->getType(), "cov") || strEQ(data->getType(), "cor")) {
+    omxMatrix *cov = omxDataCovariance(data);
+    int size = cov->rows;
+    if (strEQ(data->getType(), "cor")) size -= 1;
+    int stats = triangleLoc1(size);
+    omxMatrix *means = omxDataMeans(data);
+    if (means) stats += means->rows * means->cols;
+    return stats;
+  } else {
+    return NA_INTEGER;
+  }
+}
+
+void MVNExpectation::populateAttr(SEXP algebra)
+{
+  IntegerVector RobStat = Rcpp::wrap(numObservedStats());
+  Rf_setAttrib(algebra, Rf_install("numStats"), RobStat);
+}
