@@ -1,5 +1,5 @@
 /*
- *  Copyright 2007-2019 by the individuals mentioned in the source code history
+ *  Copyright 2007-2021 by the individuals mentioned in the source code history
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -19,8 +19,8 @@
 #include "Compute.h"
 #include "EnableWarnings.h"
 
-class omxLISRELExpectation : public omxExpectation {
-	typedef omxExpectation super;
+class omxLISRELExpectation : public MVNExpectation {
+	typedef MVNExpectation super;
 	std::vector<int> exoDataColumns; // index into omxData
 	int verbose;
 	int numExoPred;
@@ -37,7 +37,6 @@ public:
 	omxMatrix *MUX, *MUY; //Place holder matrices for building means from blocks
 	//omxMatrix *C, *P, *V, *Mns; // Other Matrices, not sure what these are for.
 	omxMatrix *slope;       // exogenous predictor slopes
-	Eigen::VectorXd exoPredMean;
 
 	int numIters; // used by omxFastRAM/LISRELInverse
 	double n;
@@ -52,12 +51,12 @@ public:
 
 	omxLISRELExpectation(omxState *st, int num) : super(st, num), numExoPred(0) {}
 	virtual ~omxLISRELExpectation();
-	virtual void init();
-  virtual void connectToData();
-	virtual void compute(FitContext *fc, const char *what, const char *how);
-	virtual void populateAttr(SEXP expectation);
-	virtual omxMatrix *getComponent(const char*);
-	virtual void getExogenousPredictors(std::vector<int> &out);
+	virtual void init() override;
+	virtual void compute(FitContext *fc, const char *what, const char *how) override;
+	virtual void populateAttr(SEXP expectation) override;
+	virtual omxMatrix *getComponent(const char*) override;
+	virtual void getExogenousPredictors(std::vector<int> &out) override;
+  virtual int numLatentVars() const override { return PS->cols; }
 
 	void studyExoPred();
 };
@@ -139,12 +138,7 @@ omxLISRELExpectation::~omxLISRELExpectation()
 
 void omxLISRELExpectation::populateAttr(SEXP algebra)
 {
-	auto oo = this;
-
-  {
-    ProtectedSEXP RnumStat(Rf_ScalarReal(omxDataDF(oo->data)));
-    Rf_setAttrib(algebra, Rf_install("numStats"), RnumStat);
-  }
+  super::populateAttr(algebra);
 
 	MxRList out;
   populateNormalAttr(algebra, out);
@@ -232,7 +226,6 @@ void omxCalculateLISRELCovarianceAndMeans(omxLISRELExpectation* oro) {
 	omxMatrix* BOT = oro->BOT;
 	omxMatrix* MUX = oro->MUX;
 	omxMatrix* MUY = oro->MUY;
-	omxMatrix *slope = oro->slope;
 	omxMatrix** args = oro->args;
 	if(OMX_DEBUG) { mxLog("Running LISREL computation in omxCalculateLISRELCovarianceAndMeans."); }
 	double oned = 1.0, zerod=0.0; //, minusOned = -1.0;
@@ -334,11 +327,6 @@ void omxCalculateLISRELCovarianceAndMeans(omxLISRELExpectation* oro) {
 				omxDGEMV(FALSE, oned, GA, KA, oned, K);
 				omxCopyMatrix(MUY, TY);
 				omxDGEMV(FALSE, oned, D, K, oned, MUY);
-				if (slope) {
-					EigenVectorAdaptor Emean(MUY);
-					EigenMatrixAdaptor Eslope(slope);
-					Emean += Eslope * oro->exoPredMean;
-				}
 			//}
 
 			/* Build means from blocks */
@@ -382,11 +370,6 @@ void omxCalculateLISRELCovarianceAndMeans(omxLISRELExpectation* oro) {
 		if(Means != NULL) {
 				omxCopyMatrix(Means, TY);
 				omxDGEMV(FALSE, oned, D, AL, oned, Means);
-				if (slope) {
-					EigenVectorAdaptor Emean(Means);
-					EigenMatrixAdaptor Eslope(slope);
-					Emean += Eslope * oro->exoPredMean;
-				}
 		}
 	}
 	if (Means && OMX_DEBUG_ALGEBRA) omxPrintMatrix(Means, "....LISREL: Model-implied Means Vector:");
@@ -580,7 +563,7 @@ void omxLISRELExpectation::init()
 	} else LISobj->means  = 	NULL;
 	//TODO: Adjust means processing to allow only Xs or only Ys
 
-	if (currentState->isClone()) {
+	if (!currentState->isTopState()) {
     auto pex = (omxLISRELExpectation*) currentState->getParent(this);
     if (pex->slope) {
       numExoPred = pex->numExoPred;
@@ -615,7 +598,9 @@ void omxLISRELExpectation::getExogenousPredictors(std::vector<int> &out)
 
 void omxLISRELExpectation::studyExoPred() // compare with similar function for RAM
 {
-	if (data->defVars.size() == 0 || !TY || !TY->isSimple() || !PS->isSimple()) return;
+	if (data->defVars.size() == 0 || !TY ||
+      TY->isAlgebra() || TY->populateDependsOnDefinitionVariables() ||
+      PS->isAlgebra() || PS->populateDependsOnDefinitionVariables()) return;
 
 	Eigen::VectorXd estSave;
 	currentState->setFakeParam(estSave);
@@ -672,15 +657,3 @@ void omxLISRELExpectation::addSlopeMatrix()
 		ex += 1;
 	}
 }
-
-void omxLISRELExpectation::connectToData()
-{
-  super::connectToData();
-
-	exoPredMean.resize(exoDataColumns.size());
-	for (int cx=0; cx < int(exoDataColumns.size()); ++cx) {
-		auto &e1 = data->rawCol( exoDataColumns[cx] );
-		Eigen::Map< Eigen::VectorXd > vec(e1.d(), data->numRawRows());
-		exoPredMean[cx] = vec.mean();
-	}
-}
