@@ -1,5 +1,5 @@
 /*
- *  Copyright 2007-2019 by the individuals mentioned in the source code history
+ *  Copyright 2007-2021 by the individuals mentioned in the source code history
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -100,7 +100,7 @@ omxMatrix* omxInitMatrix(int nrows, int ncols, unsigned short isColMajor, omxSta
 	if(om->rows == 0 || om->cols == 0) {
 		om->data = NULL;
 	} else {
-		om->data = (double*) Calloc(nrows * ncols, double);
+		om->data = (double*) R_Calloc(nrows * ncols, double);
 	}
 
 	om->algebra = NULL;
@@ -120,7 +120,7 @@ omxMatrix* omxInitMatrix(int nrows, int ncols, unsigned short isColMajor, omxSta
 static void omxFreeInternalMatrixData(omxMatrix * om)
 {
 	if(!om->owner && om->data != NULL) {
-		Free(om->data);
+		R_Free(om->data);
 	}
 	om->owner = NULL;
 	om->data = NULL;
@@ -154,7 +154,7 @@ void omxCopyMatrix(omxMatrix *dest, omxMatrix *orig) {
 	} else {
 		if(regenerateMemory) {
 			omxFreeInternalMatrixData(dest);											// Free and regenerate memory
-			dest->setData((double*) Calloc(dest->rows * dest->cols, double));
+			dest->setData((double*) R_Calloc(dest->rows * dest->cols, double));
 		}
 		if (dest->data != orig->data) {  // if equal then programmer stop? TODO
 			memcpy(dest->data, orig->data, dest->rows * dest->cols * sizeof(double));
@@ -200,7 +200,7 @@ bool omxMatrix::canDiscard()
 }
 
 void omxFreeMatrix(omxMatrix *om) {
-    
+
     if(om == NULL) return;
 
 	omxFreeInternalMatrixData(om);
@@ -215,12 +215,19 @@ void omxFreeMatrix(omxMatrix *om) {
 		om->fitFunction = NULL;
 	}
 
-	if (om->freeColnames) for (auto cn : om->colnames) free((void*)cn);
-	if (om->freeRownames) for (auto rn : om->rownames) free((void*)rn);
-	
+  om->clearDimnames();
+
 	if (!om->hasMatrixNumber) delete om;
 }
 
+void omxMatrix::clearDimnames()
+{
+	if (freeColnames) { for (auto cn : colnames) free((void*)cn); freeColnames=false; }
+	if (freeRownames) { for (auto rn : rownames) free((void*)rn); freeRownames=false; }
+  colnames.clear();
+  rownames.clear();
+}
+
 /**
  * Copies an omxMatrix to a new R matrix object
  *
@@ -264,18 +271,20 @@ omxMatrix* omxNewIdentityMatrix(int nrows, omxState* state) {
 
 omxMatrix* omxDuplicateMatrix(omxMatrix* src, omxState* newState) {
 	omxMatrix* newMat;
-    
+
 	if(src == NULL) return NULL;
 	newMat = omxInitMatrix(src->rows, src->cols, TRUE, newState);
 	omxCopyMatrix(newMat, src);
 	newMat->hasMatrixNumber = src->hasMatrixNumber;
 	newMat->matrixNumber    = src->matrixNumber;
 	newMat->nameStr = src->nameStr;
-    
+
 	newMat->rownames = src->rownames;
 	newMat->colnames = src->colnames;
 
-    return newMat;    
+  if (src->penalty) newMat->penalty = src->penalty->clone(newMat);
+
+    return newMat;
 }
 
 void omxMatrix::copyAttr(omxMatrix *src)
@@ -298,14 +307,14 @@ int omxMatrix::lookupColumnByName(const char *target)
 void omxResizeMatrix(omxMatrix *om, int nrows, int ncols)
 {
 	// Always Recompute() before you Resize().
-	if(false && OMX_DEBUG) { 
+	if(false && OMX_DEBUG) {
 		mxLog("Resizing matrix from (%d, %d) to (%d, %d)",
 			om->rows, om->cols, nrows, ncols);
 	}
 
 	if( (om->rows != nrows || om->cols != ncols)) {
 		omxFreeInternalMatrixData(om);
-		om->setData((double*) Calloc(nrows * ncols, double));
+		om->setData((double*) R_Calloc(nrows * ncols, double));
 	}
 
 	om->rows = nrows;
@@ -326,11 +335,11 @@ double* omxLocationOfMatrixElement(omxMatrix *om, int row, int col) {
 
 void vectorElementError(int index, int numrow, int numcol) {
 	if ((numrow > 1) && (numcol > 1)) {
-		mxThrow("Requested improper index (%d) from a malformed vector of dimensions (%d, %d)", 
+		mxThrow("Requested improper index (%d) from a malformed vector of dimensions (%d, %d)",
 			index, numrow, numcol);
 	} else {
 		int Rf_length = (numrow > 1) ? numrow : numcol;
-		mxThrow("Requested improper index (%d) from vector of Rf_length (%d)", 
+		mxThrow("Requested improper index (%d) from vector of Rf_length (%d)",
 			index, Rf_length);
 	}
 }
@@ -347,7 +356,7 @@ void setMatrixError(omxMatrix *om, int row, int col, int numrow, int numcol) {
 	} else {
 		typeString = matrixString;
 	}
-	mxThrow("Attempted to set row and column (%d, %d) in %s \"%s\" with dimensions %d x %d.", 
+	mxThrow("Attempted to set row and column (%d, %d) in %s \"%s\" with dimensions %d x %d.",
 		row, col, typeString, om->name(), numrow, numcol);
 }
 
@@ -358,11 +367,11 @@ void matrixElementError(int row, int col, omxMatrix *om) {
 
 void setVectorError(int index, int numrow, int numcol) {
 	if ((numrow > 1) && (numcol > 1)) {
-		mxThrow("Attempting to set improper index (%d) from a malformed vector of dimensions (%d, %d)", 
+		mxThrow("Attempting to set improper index (%d) from a malformed vector of dimensions (%d, %d)",
 			index, numrow, numcol);
 	} else {
 		int Rf_length = (numrow > 1) ? numrow : numcol;
-		mxThrow("Setting improper index (%d) from vector of Rf_length %d", 
+		mxThrow("Setting improper index (%d) from vector of Rf_length %d",
 			index, Rf_length);
 	}
 }
@@ -383,7 +392,7 @@ void omxMarkClean(omxMatrix *om)
 	}
 }
 
-omxMatrix* omxNewMatrixFromRPrimitive0(SEXP rObject, omxState* state, 
+omxMatrix* omxNewMatrixFromRPrimitive0(SEXP rObject, omxState* state,
 	unsigned short hasMatrixNumber, int matrixNumber) {
 /* Creates and populates an omxMatrix with details from an R matrix object. */
 	if (Rf_isMatrix(rObject) && Rf_length(rObject) == 1 &&
@@ -393,7 +402,7 @@ omxMatrix* omxNewMatrixFromRPrimitive0(SEXP rObject, omxState* state,
 	return omxFillMatrixFromRPrimitive(om, rObject, state, hasMatrixNumber, matrixNumber);
 }
 
-omxMatrix* omxNewMatrixFromRPrimitive(SEXP rObject, omxState* state, 
+omxMatrix* omxNewMatrixFromRPrimitive(SEXP rObject, omxState* state,
 	unsigned short hasMatrixNumber, int matrixNumber) {
 /* Creates and populates an omxMatrix with details from an R matrix object. */
 	omxMatrix *om = NULL;
@@ -521,7 +530,7 @@ void omxMatrix::unshareMemoryWithR()
 {
 	if (!owner) return;
 
-	double *copy = (double*) Realloc(NULL, rows * cols, double);
+	double *copy = (double*) R_Realloc(NULL, rows * cols, double);
 	memcpy(copy, data, rows * cols * sizeof(double));
 	setData(copy);
 	owner = NULL;
@@ -538,8 +547,8 @@ void omxToggleRowColumnMajor(omxMatrix *mat) {
 	int i, j;
 	int nrows = mat->rows;
 	int ncols = mat->cols;
-	
-	double *newdata = (double*) Calloc(nrows * ncols, double);
+
+	double *newdata = (double*) R_Calloc(nrows * ncols, double);
 	double *olddata = mat->data;
 
 	if (mat->colMajor) {
@@ -565,13 +574,13 @@ void omxTransposeMatrix(omxMatrix *mat)
 {
 	std::swap(mat->colnames, mat->rownames);
 	mat->colMajor = !mat->colMajor;
-	
+
 	if(mat->rows != mat->cols){
         int mid = mat->rows;
         mat->rows = mat->cols;
         mat->cols = mid;
 	}
-	
+
 	omxMatrixLeadingLagging(mat);
 }
 
@@ -615,7 +624,7 @@ void omxRemoveRowsAndColumns(omxMatrix *om, int rowsRemoved[], int colsRemoved[]
 
 	om->rows = newRows;
 	om->cols = newCols;
-	
+
 	if (om->colMajor) {
 		int nextCol = 0;
 		for(int j = 0; j < origCols; j++) {
@@ -664,6 +673,23 @@ void omxMatrix::markPopulatedEntries()
 	}
 }
 
+bool omxMatrix::populateDependsOnDefinitionVariables()
+{
+	if (populate.size() == 0) return false;
+	for (size_t pi = 0; pi < populate.size(); pi++) {
+		populateLocation &pl = populate[pi];
+		int index = pl.from;
+		omxMatrix* sourceMatrix;
+		if (index < 0) {
+			sourceMatrix = currentState->matrixList[~index];
+		} else {
+			sourceMatrix = currentState->algebraList[index];
+		}
+    if (sourceMatrix->dependsOnDefinitionVariables()) return true;
+  }
+  return false;
+}
+
 void omxMatrix::omxPopulateSubstitutions(int want, FitContext *fc)
 {
 	if (populate.size() == 0) return;
@@ -743,11 +769,14 @@ void omxRecompute(omxMatrix *matrix, FitContext *fc)
 	int want = matrix->currentState->getWantStage();
 	matrix->omxPopulateSubstitutions(want, fc); // could be an algebra!
 
-	if(!omxNeedsUpdate(matrix)) return;
+  // FF_COMPUTE_PREOPTIMIZE should use a different set of functions
+  // to traverse the computation graph so we can remove the
+  // conditional here.
+	if(!(want & FF_COMPUTE_PREOPTIMIZE) && !omxNeedsUpdate(matrix)) return;
 
 	if(matrix->algebra) omxAlgebraRecompute(matrix, want, fc);
 	else if(matrix->fitFunction) {
-		omxFitFunctionCompute(matrix->fitFunction, want, fc);
+    matrix->fitFunction->recompute(want, fc);
 	}
 	if (want & FF_COMPUTE_FIT) {
 		omxMarkClean(matrix);
@@ -886,7 +915,7 @@ void omxMatrixHorizCat(omxMatrix** matList, int numArgs, omxMatrix* result)
 
 	if (allArgumentsColMajor) {
 		int offset = 0;
-		for(int j = 0; j < numArgs; j++) {	
+		for(int j = 0; j < numArgs; j++) {
 			omxMatrix* current = matList[j];
 			int size = current->rows * current->cols;
 			memcpy(result->data + offset, current->data, size * sizeof(double));
@@ -934,9 +963,9 @@ void omxMatrixVertCat(omxMatrix** matList, int numArgs, omxMatrix* result)
 
 	if (allArgumentsRowMajor) {
 		int offset = 0;
-		for(int j = 0; j < numArgs; j++) {	
+		for(int j = 0; j < numArgs; j++) {
 			omxMatrix* current = matList[j];
-			int size = current->rows * current->cols;	
+			int size = current->rows * current->cols;
 			memcpy(result->data + offset, current->data, size * sizeof(double));
 			offset += size;
 		}
@@ -991,7 +1020,7 @@ int omxMatrix::numNonConstElements() const
 
 	case 2: //Full
 		return rows * cols;
-		
+
 	case 4: //Lower
 	case 7: //Symm
 		return triangleLoc1(rows);
@@ -1012,18 +1041,23 @@ int omxMatrix::numNonConstElements() const
 	}
 }
 
-void MoorePenroseInverse(Eigen::Ref<Eigen::MatrixXd> mat)
+void MoorePenroseInverseSq(Eigen::Ref<Eigen::MatrixXd> mat)
 {
-	Eigen::BDCSVD<Eigen::MatrixXd> svd(mat, Eigen::ComputeFullV | Eigen::ComputeFullU);
+  if (mat.rows() != mat.cols()) mxThrow("MoorePenroseInverseSq must be square");
+  MoorePenroseInverse(mat, mat);
+}
+
+void MoorePenroseInverse(Eigen::Ref<Eigen::MatrixXd> in, Eigen::Ref<Eigen::MatrixXd> out)
+{
+	Eigen::BDCSVD<Eigen::MatrixXd> svd(in, Eigen::ComputeFullV | Eigen::ComputeThinU);
 	Eigen::VectorXd sv = svd.singularValues();
 	for (int v1=0; v1 < sv.size(); ++v1) {
 		if (sv[v1] > 1e-6) sv[v1] = 1.0/sv[v1];
 		else sv[v1] = 0;
 	}
-	mat.derived() = svd.matrixV() * sv.asDiagonal() * svd.matrixU().transpose();
+	out.derived() = svd.matrixV() * sv.asDiagonal() * svd.matrixU().transpose();
 }
 
-
 SEXP omxMatrix::asR()
 {
 	int m = rows, n = cols;
