@@ -1,5 +1,5 @@
 /*
- *  Copyright 2007-2019 by the individuals mentioned in the source code history
+ *  Copyright 2007-2021 by the individuals mentioned in the source code history
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -27,8 +27,8 @@
  *
  **********************************************************/
 
-#ifndef _OMXMATRIX_H_
-#define _OMXMATRIX_H_
+#ifndef u_OMXMATRIX_H_
+#define u_OMXMATRIX_H_
 
 #include <R_ext/Arith.h>
 #include <memory>
@@ -37,6 +37,7 @@
 #include <Eigen/Core>
 #include <Eigen/Dense>
 #include "minicsv.h"
+#include "penalty.h"
 
 struct populateLocation {
 	int from;
@@ -44,8 +45,8 @@ struct populateLocation {
 	int destRow, destCol;
 
 	populateLocation() {};
-	populateLocation(int _from, int _srcRow, int _srcCol, int _destRow, int _destCol)
-	: from(_from), srcRow(_srcRow), srcCol(_srcCol), destRow(_destRow), destCol(_destCol) {};
+	populateLocation(int u_from, int u_srcRow, int u_srcCol, int u_destRow, int u_destCol)
+	: from(u_from), srcRow(u_srcRow), srcCol(u_srcCol), destRow(u_destRow), destCol(u_destCol) {};
 	void transpose() { std::swap(destRow, destCol); }
 };
 
@@ -73,10 +74,12 @@ class omxMatrix {
 								freeRownames(false), freeColnames(false)
 		{};
 	struct dtor;
+  void clearDimnames();
 	void setDependsOnParameters() { dependsOnParametersCache = true; };
 	void setDependsOnDefinitionVariables() { dependsOnDefVarCache = true; };
 	bool dependsOnParameters() const { return dependsOnParametersCache; };
 	bool dependsOnDefinitionVariables() const { return dependsOnDefVarCache; };
+  bool populateDependsOnDefinitionVariables();
 	bool hasPopulateSubstitutions() const { return populate.size(); };
 	void addPopulate(omxMatrix *from, int srcRow, int srcCol, int destRow, int destCol);
 	void transposePopulate();
@@ -95,6 +98,10 @@ class omxMatrix {
 	unsigned short colMajor;			// used for quick transpose
 	unsigned short hasMatrixNumber;		// is this object in the matrix or algebra arrays?
 	int matrixNumber;					// the offset into the matrices or algebras arrays
+  void setMatrixNumber(int xx) {
+    hasMatrixNumber = 1;
+    matrixNumber = xx;
+  };
 
 	SEXP owner;	// The R object owning data or NULL if we own it.
 
@@ -113,6 +120,7 @@ class omxMatrix {
 	bool canDiscard();
 	omxAlgebra* algebra;				// If it's not an algebra, this is NULL.
 	omxFitFunction* fitFunction;		// If it's not a fit function, this is NULL.
+  std::unique_ptr<Penalty> penalty;
 
 	std::string nameStr;
 	const char *name() const { return nameStr.c_str(); }
@@ -139,6 +147,8 @@ class omxMatrix {
 	void copyAttr(omxMatrix *src);
 	bool isSimple() const { return !algebra && !fitFunction && populate.size()==0; };
 	bool isAlgebra() const { return algebra != 0; }
+	bool isFitFunction() const { return fitFunction != 0; }
+	bool isPenalty() const { return penalty != 0; }
 	int numNonConstElements() const;
 	template <typename T> void loadFromStream(T &st);
 	int size() const { return rows * cols; }
@@ -201,7 +211,7 @@ struct omxMatrix::dtor {
 typedef std::unique_ptr< omxMatrix, omxMatrix::dtor > omxMatrixPtr;
 
 /* Matrix Creation Functions */
-omxMatrix* omxNewMatrixFromRPrimitive0(SEXP rObject, omxState* state, 
+omxMatrix* omxNewMatrixFromRPrimitive0(SEXP rObject, omxState* state,
        unsigned short hasMatrixNumber, int matrixNumber);
 	omxMatrix* omxNewMatrixFromRPrimitive(SEXP rObject, omxState *state,
 	unsigned short hasMatrixNumber, int matrixNumber); 							// Create an omxMatrix from an R object
@@ -227,7 +237,7 @@ void omxResizeMatrix(omxMatrix *source, int nrows, int ncols);
 	omxMatrix* omxFillMatrixFromRPrimitive(omxMatrix* om, SEXP rObject, omxState *state,
 		unsigned short hasMatrixNumber, int matrixNumber); 								// Populate an omxMatrix from an R object
 	void omxTransposeMatrix(omxMatrix *mat);												// Transpose a matrix in place.
-	void omxToggleRowColumnMajor(omxMatrix *mat);										// Transform row-major into col-major and vice versa 
+	void omxToggleRowColumnMajor(omxMatrix *mat);										// Transform row-major into col-major and vice versa
 
 /* Function wrappers that switch based on inclusion of algebras */
 	void omxPrint(omxMatrix *source, const char* d);
@@ -428,7 +438,7 @@ static OMXINLINE void omxDSYMV(double alpha, omxMatrix* mat,            // resul
 				omxMatrix* vec, double beta, omxMatrix* result) {       // only A is symmetric, and B is a vector
 	if(OMX_DEBUG) {
 		int nVecEl = vec->rows * vec->cols;
-		// mxLog("DSYMV: %c, %d, %f, 0x%x, %d, 0x%x, %d, %f, 0x%x, %d\n", u, (mat->cols),alpha, mat->data, (mat->leading), 
+		// mxLog("DSYMV: %c, %d, %f, 0x%x, %d, 0x%x, %d, %f, 0x%x, %d\n", u, (mat->cols),alpha, mat->data, (mat->leading),
 	                    // vec->data, onei, beta, result->data, onei); //:::DEBUG:::
 		if(mat->cols != nVecEl) {
 			mxThrow("Mismatch in symmetric vector/matrix multiply: %s (%d x %d) * (%d x 1).\n", "symmetric", mat->rows, mat->cols, nVecEl); // :::DEBUG:::
@@ -438,7 +448,7 @@ static OMXINLINE void omxDSYMV(double alpha, omxMatrix* mat,            // resul
 	EigenMatrixAdaptor Emat(mat);
 	EigenVectorAdaptor Evec(vec);
 	EigenVectorAdaptor Eresult(result);
-	
+
 	Eresult.derived() = alpha * (Emat.selfadjointView<Eigen::Upper>() * Evec).array() + beta;
 }
 
@@ -488,7 +498,7 @@ std::string mxStringifyMatrix(const char *name, const Eigen::DenseBase<T> &mat,
 {
 	std::string buf;
 
-	if (!debug && mat.rows() * mat.cols() > 1000) {
+	if (!debug && mat.rows() * mat.cols() > 1500) {
 		buf = string_snprintf("%s is too large to print # %dx%d\n",
 				name, mat.rows(), mat.cols());
 		return buf;
@@ -578,7 +588,8 @@ double trace_prod(const Eigen::MatrixBase<T1> &t1, const Eigen::MatrixBase<T2> &
 	return sum;
 }
 
-void MoorePenroseInverse(Eigen::Ref<Eigen::MatrixXd> mat);
+void MoorePenroseInverseSq(Eigen::Ref<Eigen::MatrixXd> mat);
+void MoorePenroseInverse(Eigen::Ref<Eigen::MatrixXd> in, Eigen::Ref<Eigen::MatrixXd> out);
 
 // https://forum.kde.org/viewtopic.php?f=74&t=96706
 // https://forum.kde.org/viewtopic.php?f=74&t=124421
@@ -618,7 +629,7 @@ template <typename T> void omxMatrix::loadFromStream(T &st)
 			}
 		}
 		break;
-		
+
 	case 4: //Lower
 		for (int cx=0; cx < cols; ++cx) {
 			for (int rx=cx; rx < rows; ++rx) {
@@ -673,4 +684,4 @@ template <typename T> void omxMatrix::loadFromStream(T &st)
 
 void MatrixInvert1(omxMatrix *target);
 
-#endif /* _OMXMATRIX_H_ */
+#endif /* u_OMXMATRIX_H_ */
