@@ -1,5 +1,5 @@
  /*
- *  Copyright 2007-2019 by the individuals mentioned in the source code history
+ *  Copyright 2007-2021 by the individuals mentioned in the source code history
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -34,14 +34,14 @@ struct omxWLSFitFunction : omxFitFunction {
 	bool fullWeight;
 
 	omxWLSFitFunction() :
-		type("WLS"), continuousType("cumulants"), fullWeight(true) {};
+		expectedFlattened(0), P(0), B(0), type("WLS"), continuousType("cumulants"), fullWeight(true) {};
 	virtual ~omxWLSFitFunction();
-	virtual void init();
-	virtual void compute(int ffcompute, FitContext *fc);
-	virtual void populateAttr(SEXP algebra);
+	virtual void init() override;
+	virtual void compute2(int ffcompute, FitContext *fc) override;
+	virtual void populateAttr(SEXP algebra) override;
 
 	void prepData();
-	virtual void invalidateCache()
+	virtual void invalidateCache() override
 	{
 		omxFreeMatrix(observedFlattened);
 		observedFlattened = 0;
@@ -59,7 +59,7 @@ omxWLSFitFunction::~omxWLSFitFunction()
 	omxFreeMatrix(owo->P);
 }
 
-void omxWLSFitFunction::compute(int want, FitContext *fc)
+void omxWLSFitFunction::compute2(int want, FitContext *fc)
 {
 	auto *oo = this;
 	auto *owo = this;
@@ -77,6 +77,8 @@ void omxWLSFitFunction::compute(int want, FitContext *fc)
 	if(OMX_DEBUG) { mxLog("Beginning WLS Evaluation.");}
 	// Requires: Data, means, covariances.
 
+  if (want & FF_COMPUTE_GRADIENT) invalidateGradient(fc);
+
 	double sum = 0.0;
 
 	omxMatrix *eCov, *eMeans, *oFlat, *eFlat;
@@ -101,10 +103,19 @@ void omxWLSFitFunction::compute(int want, FitContext *fc)
 	omxDAXPY(-1.0, eFlat, B);
 	//if(OMX_DEBUG) {omxPrintMatrix(B, "....WLS Observed - Expected Vector: "); }
 
-	omxMatrix *weights = expectation->data->getSingleObsSummaryStats().acovMat;
+  auto &oss = expectation->data->getSingleObsSummaryStats();
+	omxMatrix *weights = oss.useWeight;
 	if(weights != NULL) {
 		//if(OMX_DEBUG_ALGEBRA) {omxPrintMatrix(weights, "....WLS Weight Matrix: "); }
-		omxDGEMV(TRUE, 1.0, weights, B, 0.0, P);
+
+		if (units == FIT_UNITS_SQUARED_RESIDUAL) { // diagonally weighted
+      EigenMatrixAdaptor Ew(weights);
+      EigenVectorAdaptor EB(B);
+      EigenVectorAdaptor EP(P);
+      EP.derived() = Ew.diagonal().array() * EB.array();
+    } else {
+      omxDGEMV(TRUE, 1.0, weights, B, 0.0, P);
+    }
 	} else {
 		// ULS Case: Memcpy faster than dgemv.
 		omxCopyMatrix(P, B);
@@ -113,20 +124,22 @@ void omxWLSFitFunction::compute(int want, FitContext *fc)
 
 	sum = omxDDOT(P, B);
 
-	oo->matrix->data[0] = sum;
-
-	if(OMX_DEBUG) { mxLog("WLSFitFunction value comes to: %f.", oo->matrix->data[0]); }
+	matrix->data[0] = sum;
+  scale = oss.totalWeight;
 
+	if(OMX_DEBUG) { mxLog("WLSFitFunction %.5g %f", oo->matrix->data[0], oo->scale); }
 }
 
 void omxWLSFitFunction::populateAttr(SEXP algebra)
 {
+  if (!observedFlattened) return;
+
 	if(OMX_DEBUG) { mxLog("Populating WLS Attributes."); }
 
 	omxWLSFitFunction *argStruct = this;
 	omxMatrix *expCovInt = argStruct->expectedCov;	    		// Expected covariance
 	omxMatrix *expMeanInt = argStruct->expectedMeans;			// Expected means
-	omxMatrix *weightInt = expectation->data->getSingleObsSummaryStats().acovMat;
+	omxMatrix *weightInt = expectation->data->getSingleObsSummaryStats().useWeight;
 
 	SEXP expCovExt, expMeanExt, gradients;
 	Rf_protect(expCovExt = Rf_allocMatrix(REALSXP, expCovInt->rows, expCovInt->cols));
@@ -194,7 +207,7 @@ void omxWLSFitFunction::prepData()
 
 	omxData* dataMat = oo->expectation->data;
 
-	if (!matrix->currentState->isClone()) {
+	if (matrix->currentState->isTopState()) {
 		std::vector<int> exoPred;
 		expectation->getExogenousPredictors(exoPred);
 		// how to prohibit def vars? TODO
@@ -214,7 +227,7 @@ void omxWLSFitFunction::prepData()
 
 	omxMatrix *means = obsStat.meansMat;
 	omxMatrix *obsThresholdsMat = obsStat.thresholdMat;
-	omxMatrix *weights = obsStat.acovMat;
+	omxMatrix *weights = obsStat.useWeight;
 	std::vector< omxThresholdColumn > &oThresh = obsStat.thresholdCols;
 
 	auto &eThresh = oo->expectation->getThresholdInfo();
@@ -288,9 +301,12 @@ void omxWLSFitFunction::prepData()
 	if(OMX_DEBUG) {omxPrintMatrix(observedFlattened, "....WLS Observed Vector: "); }
 
 	/* Temporary storage for calculation */
-	expectedFlattened = omxInitMatrix(vectorSize, 1, TRUE, matrix->currentState);
-	P = omxInitMatrix(1, vectorSize, TRUE, matrix->currentState);
-	B = omxInitMatrix(vectorSize, 1, TRUE, matrix->currentState);
+	if (!expectedFlattened) expectedFlattened = omxInitMatrix(vectorSize, 1, TRUE, matrix->currentState);
+  omxResizeMatrix(expectedFlattened, vectorSize, 1);
+	if (!P) P = omxInitMatrix(1, vectorSize, TRUE, matrix->currentState);
+  omxResizeMatrix(P, 1, vectorSize);
+	if (!B) B = omxInitMatrix(vectorSize, 1, TRUE, matrix->currentState);
+  omxResizeMatrix(B, vectorSize, 1);
 }
 
 void omxWLSFitFunction::init()
@@ -325,9 +341,18 @@ void omxWLSFitFunction::init()
     expectation->connectToData();
   }
 
+  if (false) {
+    // This is too strict because we might use def vars in a way that
+    // doesn't affect the model expectation (i.e. generate a column of
+    // the interaction between two other columns).
+    omxData *data = expectation->data;
+    if (data->defVars.size()) {
+      mxThrow("%s: There are %d definition variables. "
+              "The WLS fit function cannot handle definition variables except as exogenous predictors.",
+              name(), int(data->defVars.size()));
+    }
+  }
+
 	observedFlattened = 0;
-	expectedFlattened = 0;
-	B = 0;
-	P = 0;
   canDuplicate = true;
 }
